# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.


class CrystalInfo:
    def __init__(self, nominal_hz, chirp_f0, chirp_f1, search_f0, search_f1,
                 search_df, is_valid_freq_f0, is_valid_freq_f1,
                 is_valid_peak_f0, is_valid_peak_f1, phase_shift_deg=0):
        self.nominal_hz        = nominal_hz
        self.chirp_f0          = chirp_f0
        self.chirp_f1          = chirp_f1
        self.search_f0         = search_f0
        self.search_f1         = search_f1
        self.search_min_f      = min(search_f0, search_f1)
        self.search_max_f      = max(search_f0, search_f1)
        self.search_df         = search_df
        self.is_valid_freq_f0  = is_valid_freq_f0
        self.is_valid_freq_f1  = is_valid_freq_f1
        self.is_valid_peak_f0  = is_valid_peak_f0
        self.is_valid_peak_f1  = is_valid_peak_f1
        self.phase_shift_deg   = phase_shift_deg

    def is_valid_freq(self, hz):
        return self.is_valid_freq_f0 <= hz <= self.is_valid_freq_f1

    def is_valid_peak(self, hz):
        return self.is_valid_peak_f0 <= hz <= self.is_valid_peak_f1


CRYSTAL_INFOS = {
    # Gas crystal which operates at 20 kHz and only is used to measure density
    # and viscosity of gas.  This can have a very high Q-factor so when
    # searching the frequency delta should be small.
    20000 : CrystalInfo(20000, 19000, 21000, 19000, 20500, 10, 10000, 21000,
                        15000, 20000),

    # Fluid crystal which operates at 32.768 kHz and is used to measure air for
    # baseline purposes and otherwise fluids.  In fluid, this will have a very
    # low Q-factor so when searching the frequency delta can be very large.
    32768 : CrystalInfo(32768, 28000, 33500, 20000, 35000, 50, 10000, 45000,
                        15000, 35000),

    # Pressure crystal.  This is in vacuum and typically has a very high Q-
    # factor, so when searching the frequency delta should be small.
    50000 : CrystalInfo(50000, 48000, 52000, 48000, 52000, 15, 48000, 52000,
                        48000, 52000),

    # Temperature crystal.  This is in vacuum and has a very high Q-factor, so
    # when searching the frequency delta should be small.
    262000 : CrystalInfo(262000, 260000, 264000, 260000, 264000, 15,
                         260000, 264000, 260000, 264000, -155),
}



# Copyright (c) 2022 Phase Advanced Sensor Systems, Inc.
from enum import IntEnum
import random
import time
import math
import struct

import usb
import usb.util
import btype

from xtalx.tools.iter import prange
from .cal_page import CalPage


class Status(IntEnum):
    OK              = 0
    BAD_LENGTH      = 1
    BAD_OPCODE      = 2
    QSPI_BUSY       = 3

    @staticmethod
    def rsp_to_status_str(rsp):
        try:
            s = '%s' % Status(rsp.status)
        except ValueError:
            s = '%u' % rsp.status
        return s


class Opcode(IntEnum):
    # System commands.
    GET_BOOT_STATUS          = 0x0001
    TIMESTAMP                = 0x0002
    GET_VOLTAGES             = 0x0003
    SET_AUTONOMOUS_CONFIG    = 0x0004
    START_CONTINUOUS_MODE    = 0x0005
    START_AUTONOMOUS_MODE    = 0x0006
    CLEAR_ERASE_JOURNAL      = 0x0007
    CLEAR_AUTO_JOURNAL       = 0x0008
    CLEAR_FLASH_LOG          = 0x0009
    GET_AUTONOMOUS_CONFIG    = 0x0010
    READ_FLASH_LOG           = 0x0011
    STOP_CONTINUOUS_MODE     = 0x0012

    # QSPI flash commands.
    GET_QSPI_FLASH_INFO      = 0x0100
    READ_QSPI_FLASH          = 0x0101
    START_ERASE_QSPI_FLASH   = 0x0102
    GET_QSPI_ERASE_STATUS    = 0x0103

    # Flash commands.
    FLASH_READ_CAL           = 0xBA61


class Command(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    params          = btype.Array(btype.uint32_t(), 7)
    _EXPECTED_SIZE  = 32


class Response(btype.Struct):
    opcode         = btype.uint16_t()
    tag            = btype.uint16_t()
    status         = btype.uint32_t()
    params         = btype.Array(btype.uint32_t(), 6)
    _EXPECTED_SIZE = 32


class CalPoint(btype.Struct):
    seq             = btype.uint32_t()
    vdd_mv          = btype.uint16_t()
    vbus_mv         = btype.uint16_t()
    ft              = btype.float64_t()
    fp              = btype.float64_t()
    lft             = btype.float64_t()
    lfp             = btype.float64_t()
    _EXPECTED_SIZE  = 40


class Measurement:
    def __init__(self, sensor, t_ns, seq, vdd_mv, vbus_mv, ft, fp, lft, lfp):
        self.sensor = sensor
        self.seq    = seq
        self.t_ns   = t_ns
        self.vdd    = vdd_mv / 1000
        self.vbus   = vbus_mv / 1000
        self.ft     = ft if not math.isnan(ft) else 0
        self.fp     = fp if not math.isnan(fp) else 0
        self.lft    = lft if not math.isnan(lft) else 0
        self.lfp    = lfp if not math.isnan(lfp) else 0

        self.pressure_psi = None
        self.lores_pressure_psi = None
        if sensor.poly_psi is not None:
            if self.ft and self.fp:
                self.pressure_psi = sensor.poly_psi(self.fp, self.ft)
            if self.lft and self.lfp:
                self.lores_pressure_psi = sensor.poly_psi(self.lfp, self.lft)

        self.temp_c = None
        self.lores_temp_c = None
        if sensor.poly_temp is not None:
            if self.ft:
                self.temp_c = sensor.poly_temp(self.ft)
            if self.lft:
                self.lores_temp_c = sensor.poly_temp(self.lft)

    @staticmethod
    def from_cal_point(sensor, t_ns, cp):
        return Measurement(sensor, t_ns, cp.seq, cp.vdd_mv, cp.vbus_mv, cp.ft,
                           cp.fp, cp.lft, cp.lfp)

    def to_influx_point(self, time_ns=None, measurement='xtalx_data',
                        fields=None):
        fields = fields or {}
        p = {
            'measurement' : measurement,
            'time'        : time_ns or self.t_ns,
            'tags'        : {'sensor' : self.sensor.serial_num},
            'fields'      :
            {
                'temp_freq_hz'           : float(self.ft),
                'pressure_freq_hz'       : float(self.fp),
                'lores_temp_freq_hz'     : float(self.lft),
                'lores_pressure_freq_hz' : float(self.lfp),
                **fields,
            },
        }
        if self.pressure_psi is not None:
            fields['pressure_psi'] = float(self.pressure_psi)
        if self.temp_c is not None:
            fields['temp_c'] = self.temp_c
        if self.lores_pressure_psi is not None:
            fields['lores_pressure_psi'] = float(self.lores_pressure_psi)
        if self.lores_temp_c is not None:
            fields['lores_temp_c'] = self.lores_temp_c
        return p

    def to_combined_stsdb_point(self, time_ns=None):
        return {
            'time_ns'                : time_ns or self.t_ns,
            'pressure_psi'           : self.pressure_psi,
            'temp_c'                 : self.temp_c,
            'pressure_freq_hz'       : self.fp,
            'temp_freq_hz'           : self.ft,
            'lores_pressure_psi'     : self.lores_pressure_psi,
            'lores_temp_c'           : self.lores_temp_c,
            'lores_pressure_freq_hz' : self.lfp,
            'lores_temp_freq_hz'     : self.lft,
        }


class CommandException(Exception):
    def __init__(self, rsp, rx_data):
        super().__init__(
            'Command exception: %s (%s)' % (Status.rsp_to_status_str(rsp), rsp))
        self.rsp = rsp
        self.rx_data = rx_data


class BootStatus:
    def __init__(self, post_result, pll_freq, hse_freq, ls_ticks, page_size):
        self.post_result    = post_result
        self.pll_freq       = pll_freq
        self.hse_freq       = hse_freq
        self.ls_ticks       = ls_ticks
        self.ls_freq        = pll_freq * 1024 / ls_ticks if ls_ticks else 0
        self.page_size      = page_size

    def __str__(self):
        return ('BootStatus(post_result=0x%08X, pll_freq=%u, hse_freq=%u, '
                'ls_ticks=%u)' %
                (self.post_result, self.pll_freq, self.hse_freq, self.ls_ticks))


class Voltages:
    def __init__(self, vdd_mv, vbus_mv):
        self.vdd_mv  = vdd_mv
        self.vbus_mv = vbus_mv

    def __str__(self):
        return 'Voltages(vdd_mv=%u, vbus_mv=%u)' % (self.vdd_mv, self.vbus_mv)


class AutonomousConfig:
    def __init__(self, interval_secs, flags):
        self.interval_secs = interval_secs
        self.flags         = flags


class QSPIInfo:
    def __init__(self, dev_id, m_id, d_id, write_index, nslots):
        self.dev_id      = dev_id
        self.m_id        = m_id
        self.d_id        = d_id
        self.write_index = write_index
        self.nslots      = nslots
        self.capacity    = (1 << (d_id & 0xFF))


class QSPIEraseStatus:
    def __init__(self, erase_in_progress, extended_read_register):
        self.erase_in_progress      = erase_in_progress
        self.extended_read_register = extended_read_register


class XMHTI:
    CMD_EP = 0x01
    RSP_EP = 0x82
    CAL_EP = 0x83

    def __init__(self, usb_dev):
        self.usb_dev      = usb_dev
        self.tag          = random.randint(0, 65535)
        self.last_time_ns = 0

        try:
            self.serial_num = usb_dev.serial_number
            self.git_sha1   = usb.util.get_string(usb_dev, 6)
            self.fw_version = usb_dev.bcdDevice
        except ValueError as e:
            if str(e) == 'The device has no langid':
                raise Exception(
                    'Device has no langid, ensure running as root!') from e

        self.fw_version_str = '%u.%u.%u' % (
                (self.fw_version >> 8) & 0xFF,
                (self.fw_version >> 4) & 0x0F,
                (self.fw_version >> 0) & 0x0F)
        assert self.fw_version >= 0x092

        self._synchronize()
        self._boot_status = self.get_boot_status()

        cp0_data, cp1_data = self.read_calibration_pages_raw()
        self._cal_pages = [CalPage.unpack(cp0_data), CalPage.unpack(cp1_data)]
        if self._cal_pages[0].is_valid():
            self.cal_page = self._cal_pages[0]
        elif self._cal_pages[1].is_valid():
            self.cal_page = self._cal_pages[1]
        else:
            self.cal_page = None

        self.report_id = None
        self.poly_psi  = None
        self.poly_temp = None
        if self.cal_page is not None:
            self.report_id = self.cal_page.get_report_id()
            self.poly_psi, self.poly_temp = self.cal_page.get_polynomials()

    def __str__(self):
        return 'XMHTI(%s)' % self.serial_num

    def _exec_command(self, opcode, params=None, bulk_data=b'', timeout=1000,
                      rx_len=0):
        if not params:
            params = [0, 0, 0, 0, 0, 0, 0]
        elif len(params) < 7:
            params = params + [0]*(7 - len(params))

        tag      = self.tag
        self.tag = (self.tag + 1) & 0xFFFF

        cmd  = Command(opcode=opcode, tag=tag, params=params)
        data = cmd.pack()
        l    = self.usb_dev.write(self.CMD_EP, data + bulk_data,
                                  timeout=timeout)
        assert l == 32 + len(bulk_data)

        data = self.usb_dev.read(self.RSP_EP, Response._STRUCT.size + rx_len,
                                 timeout=timeout)
        assert len(data) >= Response._STRUCT.size
        rsp = Response.unpack(data[:Response._STRUCT.size])
        assert rsp.opcode == opcode
        assert rsp.tag    == tag

        rx_data = bytes(data[Response._STRUCT.size:])
        if rsp.status != Status.OK:
            raise CommandException(rsp, rx_data)

        assert len(rx_data) == rx_len
        return rsp, rx_data

    def _exec_qspi_read_command(self, opcode, index, nslots, timeout=1000):
        tag      = self.tag
        self.tag = (self.tag + 1) & 0xFFFF
        cmd      = Command(opcode=opcode, tag=tag,
                           params=[index, nslots, 0, 0, 0, 0, 0])
        data     = cmd.pack()
        self.usb_dev.write(self.CMD_EP, data, timeout=timeout)

        # Read the data and check for errors.  TODO: We can't discriminate
        # between reading 4 slots and reading a 32-byte error response.
        data = self.usb_dev.read(self.RSP_EP, nslots * 8, timeout=timeout)
        if len(data) != nslots * 8:
            assert len(data) == Response._STRUCT.size
            rsp = Response.unpack(data)
            assert rsp.tag == tag
            raise CommandException(rsp, b'')

        return bytes(data)

    def _set_configuration(self, bConfigurationValue):
        cfg = None
        try:
            cfg = self.usb_dev.get_active_configuration()
        except usb.core.USBError as e:
            if e.strerror != 'Configuration not set':
                raise

        if cfg is None or cfg.bConfigurationValue != bConfigurationValue:
            usb.util.dispose_resources(self.usb_dev)
            self.usb_dev.set_configuration(bConfigurationValue)

    def _synchronize(self):
        self._set_configuration(0x51)
        while True:
            try:
                self.usb_dev.read(self.RSP_EP, 16384, timeout=100)
            except usb.core.USBTimeoutError:
                break
        self.stop_continuous_mode()

    def read_calibration_pages_raw(self):
        '''
        Returns the raw data bytes for each of the two calibration pages stored
        in flash.  Returns a tuple:

            (cal_data_0, cal_data_1)
        '''
        _, cal_data_0 = self._exec_command(Opcode.FLASH_READ_CAL, [0],
                                           rx_len=2048)
        _, cal_data_1 = self._exec_command(Opcode.FLASH_READ_CAL, [1],
                                           rx_len=2048)
        return cal_data_0, cal_data_1

    def read_calibration_pages(self):
        '''
        Returns CalPage structs for each calibration page in the sensor flash,
        even if the page(s) are missing or corrupt.
        '''
        cp0_data, cp1_data = self.read_calibration_pages_raw()
        cp0 = CalPage.unpack(cp0_data)
        cp1 = CalPage.unpack(cp1_data)
        return cp0, cp1

    def read_valid_calibration_page(self):
        '''
        Returns CalPage struct from the sensor flash.  Returns None if the
        calibration is not present or both pages are corrupted.
        '''
        cp0, cp1 = self.read_calibration_pages()
        return cp0 if cp0.is_valid() else cp1 if cp1.is_valid() else None

    def get_boot_status(self):
        rsp, _ = self._exec_command(Opcode.GET_BOOT_STATUS)
        return BootStatus(rsp.params[0], rsp.params[1], rsp.params[2] * 1000000,
                          rsp.params[3], rsp.params[4])

    def record_timestamp(self, unix_time=None):
        if unix_time is None:
            unix_time = int(time.time())
        self._exec_command(Opcode.TIMESTAMP, [unix_time])
        return unix_time

    def get_voltages(self):
        rsp, _ = self._exec_command(Opcode.GET_VOLTAGES)
        return Voltages(rsp.params[0], rsp.params[1])

    def set_autonomous_config(self, sample_interval_secs, flags):
        self._exec_command(Opcode.SET_AUTONOMOUS_CONFIG,
                           [sample_interval_secs, flags])

    def get_autonomous_config(self):
        rsp, _ = self._exec_command(Opcode.GET_AUTONOMOUS_CONFIG)
        return AutonomousConfig(rsp.params[0], rsp.params[1])

    def start_continuous_mode(self):
        self._exec_command(Opcode.START_CONTINUOUS_MODE)

    def stop_continuous_mode(self):
        self._exec_command(Opcode.STOP_CONTINUOUS_MODE)
        while True:
            try:
                self.usb_dev.read(self.CAL_EP, 16384, timeout=100)
            except usb.core.USBTimeoutError:
                break

    def start_autonomous_mode(self):
        self._exec_command(Opcode.START_AUTONOMOUS_MODE)

    def clear_erase_journal(self):
        self._exec_command(Opcode.CLEAR_ERASE_JOURNAL)

    def clear_auto_journal(self):
        self._exec_command(Opcode.CLEAR_AUTO_JOURNAL)

    def read_flash_log(self):
        tag      = self.tag
        self.tag = (self.tag + 1) & 0xFFFF
        cmd      = Command(opcode=Opcode.READ_FLASH_LOG, tag=tag)
        data     = cmd.pack()
        self.usb_dev.write(self.CMD_EP, data, timeout=1000)

        # Read page-sized chunks until we get a zero-length one to terminate
        # the transfer.
        data = b''
        while True:
            chunk = self.usb_dev.read(self.RSP_EP, self._boot_status.page_size,
                                      timeout=1000)
            if len(chunk) == Response._STRUCT.size:
                rsp = Response.unpack(chunk)
                assert rsp.tag == tag
                raise CommandException(rsp, b'')
            if len(chunk) == 0:
                break

            assert len(chunk) == self._boot_status.page_size
            data += chunk

        return data

    def clear_flash_log(self):
        self._exec_command(Opcode.CLEAR_FLASH_LOG)

    def get_qspi_flash_info(self):
        rsp, _ = self._exec_command(Opcode.GET_QSPI_FLASH_INFO)
        return QSPIInfo(rsp.params[0], rsp.params[1], rsp.params[2],
                        rsp.params[3], rsp.params[4])

    def read_qspi_flash(self, index, nslots):
        data = b''
        for i in prange(index, index + nslots, 16384):
            n       = min(nslots, 16384)
            data   += self._exec_qspi_read_command(Opcode.READ_QSPI_FLASH, i, n,
                                                   timeout=1000)
            nslots -= n
        return data

    def start_erase_qspi_flash(self):
        self._exec_command(Opcode.START_ERASE_QSPI_FLASH)

    def get_qspi_erase_status(self):
        rsp, _ = self._exec_command(Opcode.GET_QSPI_ERASE_STATUS)
        return QSPIEraseStatus(rsp.params[0], rsp.params[1])

    def _read_measurement(self, timeout=10):
        '''
        Reads a single block of 100 measurements.
        '''
        data    = b''
        rem     = 4008
        while rem:
            try:
                new_data = self.usb_dev.read(self.CAL_EP, rem, timeout=timeout)
                rem     -= len(new_data)
                data    += new_data
            except usb.core.USBTimeoutError:
                time.sleep(0.02)

        return data

    def read_measurements(self):
        '''
        Reads blocks of 100 measurements.  This uses a very aggressive timeout
        which breaks the read up into multiple shorter operations and seems to
        generate less USB noise on the XMHTI5.
        '''
        data = self._read_measurement(timeout=1)
        t_ns = time.time_ns()
        if t_ns <= self.last_time_ns:
            t_ns = self.last_time_ns + 1
        n_queued = struct.unpack('<I', data[:4])[0]
        assert ((len(data) - 8) % CalPoint._STRUCT.size) == 0
        N = (len(data) - 8) // CalPoint._STRUCT.size
        assert N == 100
        ms = [Measurement.from_cal_point(
                    self,
                    t_ns + (i - 100 - n_queued) * 100000000,
                    CalPoint.unpack_from(data, 8 + i * CalPoint._STRUCT.size)
                ) for i in range(N)]

        self.last_time_ns = ms[-1].t_ns

        if n_queued >= 100:
            return ms + self.read_measurements()

        return ms



# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import time

import xtalx.tools.serial

from . import bus
from . import modbus_crc


class Bus(bus.Bus):
    '''
    Modbus is a trash protocol.  It's dumbfounding that a quarter of the way
    through the 21st century, people still want to use this decades-old insane
    protocol to try and talk to their devices.  But they do, so here we are.

    Modbus delimits requests and responses using a packet structure called a
    protocol data unit (PDU).  Modbus RTU, which is the main protocol used to
    talk to Modbus devices over a serial port, delimits the PDU using idle time
    on the serial bus.  For baud rates above 19200, an inter-character delay of
    over 750us is flagged as an error condition, and an idle time of 1.75ms is
    used to delimit the end of a PDU.

    Compliant Modbus target devices will watch for these gaps between
    characters and use them to delimit the start and end of PDUs being
    transmitted on the bus (either from some other target device or from the
    master device).  When responding to a request, a compliant Modbus device
    will ensure that the characters it transmits don't have any gaps between
    them and that the end of the PDU is marked by 1.75ms of silence.  For a
    target implementation on some MCU that is running on bare metal or with
    some sort of real-time or DMA facilities, transmitting the PDU without any
    gaps is generally easy to accomplish.

    On the master side, we may be running a Python script (such as this one)
    and using a generic USB-to-RS485 adapter to access the serial bus.  The
    host-side drivers will have buffers in the kernel and the adapter itself
    will probably also have buffers or FIFOs.  When the host transmits a PDU on
    the bus, as long as the PDU is fairly short (say, 32 bytes or less) then it
    has a pretty good chance of being transmitted without any gaps - it will
    fit inside a single USB packet and so will probably fit inside any FIFO on
    the adapter side and the hardware will then push it all out to the serial
    port as quickly as possible without any gaps.

    Things get hairy when we have to receive.  Instead of a chunk of data in a
    USB packet, the adapter is seeing characters trickle in on the serial bus.
    The adapter may choose to transmit some of those characters to the host
    right away, and then it may choose to buffer some of them for later.  Or,
    maybe the host polls the adapter once and then not again for awhile since
    it is busy servicing some other device on the USB bus (or busy doing
    something completely unrelated and doesn't have time for USB at all).  So,
    the host side sees a few bytes and then there is a gap - even though the
    target device transmitted things according to spec and the bytes are just
    sitting in a buffer somewhere waiting for someone to get to them.  If that
    gap is over 1.75ms in length, then a strictly-compliant host is going to
    interpret that as a PDU delimiter and try to process a truncated packet.

    All of this is pretty dumb.  Modbus is a master-slave request-response
    protocol.  The slave can't arbitrarily transmit and once a slave is
    addressed, no other device can transmit either.  We generally just use
    commands to read a fixed number of registers, so we know exactly how long
    of a response to expect (assuming no exception is returned from the device).
    There are some commands, such as the Read Identification command, which can
    return variable-length data, but never more than 252 bytes: including
    address, function code and 16-bit CRC, the maximum length of a Modbus RTU
    frame is 256 bytes.

    This Modbus host implementation drastically relaxes the rules when
    receiving data on the serial port.  We use a 100ms (default, but
    configurable) inter-character timeout as an end-of-frame delimiter for
    those variable-length frames.  For a frame where we know the response size
    ahead of time (such as Read Holding Registers) then we end the frame after
    exactly the expected number of bytes has been received, also detecting an
    exception frame and terminating the frame appropriately if one is received.
    '''
    def __init__(self, intf, baud_rate, parity='E', **kwargs):
        super().__init__()
        self.serial = xtalx.tools.serial.from_intf(intf, baudrate=baud_rate,
                                                   parity=parity, **kwargs)

    def _read_until_gap(self, prev_data):
        data = b''
        while True:
            byte = self.serial.read(1)
            if byte == b'':
                return data

            data += byte
            if len(data) > 256:
                raise bus.ResponseOverflowException(prev_data + data)

    def _process_response(self, slave_addr, function_code, nbytes, data):
        # Check the CRC.
        expected_crc = modbus_crc.compute_as_bytes(data[:-2])
        if expected_crc != data[-2:]:
            # We've received garbage.  In the case of a fixed-length response,
            # we haven't waited for an inter-frame gap yet, so do that now
            # allowing us to drain data from a babbling device.
            if nbytes is not None:
                data += self._read_until_gap(data)
            raise bus.BadCRCException(data, expected_crc)

        # The CRC was good, but we could still have logical issues with the
        # packet.  Make sure it is from the right address and has the right
        # function code.
        if data[0] != slave_addr:
            raise bus.BadAddressException(data)
        if data[1] & 0x7F != function_code:
            raise bus.BadFunctionException(data)
        if data[1] & 0x80:
            raise bus.ExceptionResponseException(data, data[2])

        # The response makes logical sense, we are golden.
        return data

    def _read_response(self, slave_addr, function_code, response_time_ms,
                       nbytes=None):
        '''
        Read the response from a target.  If nbytes is None, the response is
        variable-length and we use the inter-frame gap with an 0.1-second
        timeout to find it.  If nbytes is not None, it is the length of the
        response, not including the address byte or CRC byte, but including
        the function code and all data that follows.
        '''
        self.serial.timeout = response_time_ms / 1000
        data = b''

        # Read the address and function code bytes, bailing if there is a
        # timeout.
        data = self.serial.read(2)
        if len(data) < 2:
            raise bus.ResponseTimeoutException(data)

        # If we got an exception function code, switch to reading a fixed-
        # length exception response.
        if data[1] & 0x80:
            nbytes = 2

        # Read the remaining data.
        self.serial.timeout = 0.1
        if nbytes is not None:
            # We have the start of a fixed-length response that looks like what
            # we expect.  Read the remaining bytes.
            data += self.serial.read(nbytes + 1)
            if len(data) != nbytes + 3:
                raise bus.ResponseTimeoutException(data)
        else:
            # We have the start of a variable-length response that looks like
            # what we expect.  Read until we hit the inter-frame gap.
            data += self._read_until_gap(data)

        return self._process_response(slave_addr, function_code, nbytes, data)

    def _send_request(self, slave_addr, data):
        '''
        Send a request to the target slave address.  The request data starts at
        the function code field; we will insert the address at the start and
        append a valid CRC here.

        Before transmitting, we have to be sure that at least 1.75ms has
        elapsed since the most recent bus activity.  For instance, a compliant
        target device will ignore any request from us that starts immediately
        after it sent its most-recent response to us until it sees 1.75ms of
        idle time.  Since we are the only master, the easiest way to do that is
        to just always sleep here first.  We use a 2ms timeout just to be safe.
        '''
        time.sleep(0.002)
        data  = bytes([slave_addr]) + data
        data += modbus_crc.compute_as_bytes(data)
        self.serial.write(data)

    def transact(self, addr, data, response_time_ms, nbytes=None):
        self._send_request(addr, data)
        return self._read_response(addr, data[0], response_time_ms,
                                   nbytes=nbytes)



# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
from .bus import (
                  BabbleException,
                  BadAddressException,
                  BadCRCException,
                  BadFunctionException,
                  Bus,
                  DeviceIDObject,
                  ExceptionResponseException,
                  ModbusException,
                  ResponseException,
                  ResponseFramingErrorException,
                  ResponseInterruptedException,
                  ResponseOverflowException,
                  ResponseUnderflowException,
                  ResponseSyntaxException,
                  ResponseTimeoutException,
                  )


__all__ = [
    'BabbleException',
    'BadAddressException',
    'BadCRCException',
    'BadFunctionException',
    'Bus',
    'DeviceIDObject',
    'ExceptionResponseException',
    'ModbusException',
    'ResponseException',
    'ResponseFramingErrorException',
    'ResponseInterruptedException',
    'ResponseOverflowException',
    'ResponseUnderflowException',
    'ResponseSyntaxException',
    'ResponseTimeoutException',
]



#!/usr/bin/env python3
# Copyright (c) 2020-2024 by Phase Advanced Sensor Systems Corp.
import argparse

import xtalx.p_sensor


def main(args):
    d = xtalx.p_sensor.find_one_xti(serial_number=args.serial_number)
    x = xtalx.p_sensor.make_xti(d)
    if args.disable_pid:
        print('Disabling PID control...')
        x._disable_pid()
    if (args.p is not None or args.i is not None or args.d is not None or
            args.setpoint_c is not None):
        print('Setting PID coefficients and setpoint...')
        x._set_pid_info(Kp=args.p, Ki=args.i, Kd=args.d,
                        setpoint_c=args.setpoint_c)

    pid_info = x._get_pid_info()
    print('Current PID configuration:')
    print('   P        = %s' % pid_info.Kp)
    print('   I        = %s' % pid_info.Ki)
    print('   D        = %s' % pid_info.Kd)
    print('   Setpoint = %s C' % pid_info.setpoint_c)

    if args.enable_pid:
        print('Enabling PID control...')
        x._enable_pid()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--serial-number', '-s')
    parser.add_argument('--enable-pid')
    parser.add_argument('--disable-pid')
    parser.add_argument('-p', type=float)
    parser.add_argument('-i', type=float)
    parser.add_argument('-d', type=float)
    parser.add_argument('--setpoint-c', type=float)
    parser.add_argument('--verbose', '-v', action='store_true')
    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2023-2025 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import threading
import argparse
import math
import time

import glotlib
import glfw

import xtalx.p_sensor
import xtalx.modbus_adapter
import xtalx.tools.modbus.serial


LINE_WIDTH = 1
RUNNING = True
TARE = 0


class TrackerWindow(glotlib.Window):
    def __init__(self, name):
        super().__init__(900, 700, msaa=4, name=name)

        self.data_gen  = -1
        self.plot_gen  = -1
        self.data_lock = threading.Lock()
        self.new_data  = []
        self.last_diff = 0

        self.p_plot = self.add_plot(
            311, limits=(-0.1, -0.5, 120, 300), max_v_ticks=10)
        self.p_lines_0 = self.p_plot.add_steps(X=[], Y=[], width=LINE_WIDTH)
        self.p_lines_1 = self.p_plot.add_steps(X=[], Y=[], width=LINE_WIDTH)
        self.p_plot.set_y_label('PSI')

        self.diff_plot = self.add_plot(
            312, limits=(-0.1, -3, 120, 3), max_v_ticks=10,
            sharex=self.p_plot)
        self.diff_lines = self.diff_plot.add_lines(X=[], Y=[], width=LINE_WIDTH)
        self.diff_plot.set_y_label('PSI')

        self.t_plot = self.add_plot(
            313, limits=(-0.1, 10, 180, 50), max_v_ticks=10,
            sharex=self.p_plot)
        self.t_lines_0 = self.t_plot.add_steps(X=[], Y=[],  width=LINE_WIDTH)
        self.t_lines_1 = self.t_plot.add_steps(X=[], Y=[],  width=LINE_WIDTH)
        self.t_plot.set_y_label('Temp (C)')

        self.pos_label = self.add_label((0.99, 0.01), '', anchor='SE')

        self.mouse_vlines = [self.p_plot.add_vline(0, color='#80C080'),
                             self.diff_plot.add_vline(0, color='#80C080'),
                             self.t_plot.add_vline(0, color='#80C080')]

        label_font = glotlib.fonts.vera_bold(48, 0)
        self.psi_label  = self.add_label(self.p_plot.bounds[2:4], '',
                                         anchor='NE', font=label_font)
        self.diff_label = self.add_label(self.diff_plot.bounds[2:4], '',
                                         anchor='NE', font=label_font)
        self.temp_label = self.add_label(self.t_plot.bounds[2:4], '',
                                         anchor='NE', font=label_font)

    def handle_mouse_moved(self, x, y):
        data_x, _ = self.p_plot._window_to_data(x, y)
        for vline in self.mouse_vlines:
            vline.set_x_data(data_x)
        self.mark_dirty()

    def handle_key_press(self, key):
        global TARE
        if key == glfw.KEY_T:
            TARE = self.last_diff
        elif key == glfw.KEY_U:
            TARE = 0

    def update_geometry(self, _t):
        updated = False

        _, _, _, data_x, data_y = self.get_mouse_pos()
        if data_x is not None:
            updated |= self.pos_label.set_text('%.10f  %.10f' %
                                               (data_x, data_y))

        new_data = None
        with self.data_lock:
            if self.new_data:
                new_data = self.new_data
                self.new_data = []

        if new_data:
            updated = True

            X  = [m[0]._timestamp for m in new_data]
            Y0 = [m[0].pressure_psi for m in new_data]
            Y1 = [m[1].pressure_psi for m in new_data]

            # Hi-res pressure measurements.
            self.p_lines_0.append_x_y_data(X, Y0)
            self.p_lines_1.append_x_y_data(X, Y1)
            self.psi_label.set_text('%.4f PSI\n%.4f PSI' %
                                    (new_data[-1][0].pressure_psi,
                                     new_data[-1][1].pressure_psi))

            # Difference measurement.
            dY = [y0 - y1 - TARE for y0, y1 in zip(Y0, Y1)]
            self.diff_lines.append_x_y_data(X, dY)
            self.diff_label.set_text('%.4f PSI' % dY[-1])
            self.last_diff = Y0[-1] - Y1[-1]

            # Hi-res temperature measurements.
            self.t_lines_0.append_x_y_data(
                X, [m[0].temp_c for m in new_data])
            self.t_lines_1.append_x_y_data(
                X, [m[1].temp_c for m in new_data])
            self.temp_label.set_text('%.4f \u00B0C\n%.4f \u00B0C' %
                                     (new_data[-1][0].temp_c,
                                      new_data[-1][1].temp_c))

        return updated

    def measurement_callback(self, m0, m1):
        with self.data_lock:
            self.new_data.append((m0, m1))
            self.mark_dirty()


def sleep_until(t0, dt=0):
    '''
    Sleeps until time (t0 + dt).
    '''
    while True:
        rem = dt - (time.time() - t0)
        if rem <= 0:
            return
        time.sleep(rem)


def measure_thread(xs, tw, csv_file, poll_rate_secs):
    t0 = time.time()
    while RUNNING:
        t = time.time()
        m0 = xs[0].read_measurement()
        m1 = xs[1].read_measurement()
        m0._timestamp = m1._timestamp = dt = t - t0
        tw.measurement_callback(m0, m1)

        if csv_file:
            temp_c0 = m0.temp_c if m0.temp_c is not None else math.nan
            pressure_psi0 = (m0.pressure_psi if m0.pressure_psi is not None
                             else math.nan)
            temp_c1 = m1.temp_c if m1.temp_c is not None else math.nan
            pressure_psi1 = (m1.pressure_psi if m1.pressure_psi is not None
                             else math.nan)
            diff_psi = pressure_psi0 - pressure_psi1
            csv_file.write('%.6f,%.6f,%.5f,%.5f,%.5f,%.2f,%.5f,%.2f,%.5f\n' % (
                t, dt, diff_psi, TARE, diff_psi - TARE, temp_c0, pressure_psi0,
                temp_c1, pressure_psi1))
            csv_file.flush()

        sleep_until(t, poll_rate_secs)


def make_sensors(args):
    dev = xtalx.modbus_adapter.find_one_mba()
    if dev is None:
        raise Exception('No matching devices.')

    bus = xtalx.modbus_adapter.make_mba(dev, baud_rate=args.baud_rate)
    bus.set_vext(True)
    time.sleep(0.1)
    return [xtalx.p_sensor.XHTISM(bus, int(addr, 0))
            for addr in args.modbus_addr]


def main(args):
    global RUNNING

    if len(args.modbus_addr) != 2:
        print('Must specify exactly two sensors.')
        return

    xs = make_sensors(args)

    if args.csv_file:
        hdr = 'Time (s)'
        hdr += ',Delta Time (s)'
        hdr += ',Pressure Difference (PSI)'
        hdr += ',Tare Value (PSI)'
        hdr += ',Tared Pressure Difference (PSI)'
        hdr += ',' + xs[0].serial_num + ' Temp (C)'
        hdr += ',' + xs[0].serial_num + ' Pressure (PSI)'
        hdr += ',' + xs[1].serial_num + ' Temp (C)'
        hdr += ',' + xs[1].serial_num + ' Pressure (PSI)'
        hdr += '\n'

        csv_file = open(  # pylint: disable=R1732
            args.csv_file, 'a+', encoding='utf8')

        pos = csv_file.tell()
        if pos != 0:
            csv_file.seek(0)
            if csv_file.read(len(hdr)) != hdr:
                print('%s does not appear to be a difference log file.' %
                      args.csv_file)
                return
            csv_file.seek(pos)
        else:
            csv_file.write(hdr)
            csv_file.flush()
    else:
        csv_file = None

    tw  = TrackerWindow(xs[0].serial_num + ' - ' + xs[1].serial_num)
    mt  = threading.Thread(target=measure_thread,
                           args=(xs, tw, csv_file, float(args.poll_rate_secs)))
    mt.start()

    try:
        glotlib.interact()
    except KeyboardInterrupt:
        print()
    finally:
        RUNNING = False
        mt.join()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--baud-rate', type=int, default=115200)
    parser.add_argument('--modbus-addr', '-m', action='append')
    parser.add_argument('--csv-file')
    parser.add_argument('--poll-rate-secs', default=0)
    args = parser.parse_args()
    main(args)


if __name__ == '__main__':
    _main()



# Copyright (c) 2025-2026 Phase Advanced Sensor Systems, Inc.
from enum import IntEnum

import xtalx.usbcmd


class Feature(IntEnum):
    CURRENT_MEASUREMENT     = (1 << 0)


class Status(IntEnum):
    XACT_FAILED     = 7

    @staticmethod
    def rsp_to_status_str(rsp):
        try:
            return '%s' % Status(rsp.status)
        except ValueError:
            return xtalx.usbcmd.Status.rsp_to_status_str(rsp)


class Opcode(IntEnum):
    SET_VEXT                 = 0x0001
    SET_FREQUENCY            = 0x0002
    XACT                     = 0x0003
    MEASURE_CURRENT          = 0x0004
    JUMP_BOOTLOADER          = 0x7B42


CMD_EP = 0x01
RSP_EP = 0x82


class SPIA(xtalx.usbcmd.Device):
    def __init__(self, usb_dev):
        super().__init__(usb_dev, CMD_EP, RSP_EP, 256, 256,
                         git_sha1_index=6,
                         default_configuration=0x84)

        self.features = 0
        if self.serial_num.startswith('XSPI-2'):
            if self.fw_version >= 0x091:
                self.features |= Feature.CURRENT_MEASUREMENT

    def __str__(self):
        return 'SPIA(%s)' % self.serial_num

    def set_vext(self, enabled):
        return self._exec_command(Opcode.SET_VEXT, [enabled])

    def set_spi_frequency(self, hz):
        rsp, _ = self._exec_command(Opcode.SET_FREQUENCY, [hz])
        return rsp.params[0]

    def measure_current(self):
        if (self.features & Feature.CURRENT_MEASUREMENT) == 0:
            return 0.0

        rsp, _ = self._exec_command(Opcode.MEASURE_CURRENT)
        adc_iir = (((rsp.params[1] << 32) & 0xFFFFFFFF00000000) |
                   ((rsp.params[0] <<  0) & 0x00000000FFFFFFFF))
        return (3.3 / (4096 * 1000 * 0.033 * 2**52)) * adc_iir

    def enter_dfu_mode(self):
        return self._exec_command(Opcode.JUMP_BOOTLOADER, [0xA47B39FE])

    def transact(self, data):
        rsp, rx_data = self._exec_command(Opcode.XACT, data=data)

        assert rsp.data_len == len(data)
        assert rsp.data_len == len(rx_data)

        return rx_data



# Copyright (c) 2025-2026 Phase Advanced Sensor Systems, Inc.
from enum import IntEnum
import platform
import random
import time

import usb
import usb.util
import btype


class Status(IntEnum):
    OK              = 0
    ABORTED         = 1
    BAD_OPCODE      = 2
    BAD_CMD_LENGTH  = 3
    BAD_DATA_LENGTH = 4
    BAD_ADDR        = 5
    BAD_PARAM       = 6

    @staticmethod
    def rsp_to_status_str(rsp):
        try:
            s = '%s' % Status(rsp.status)
        except ValueError:
            s = '%u' % rsp.status
        return s


class Command(btype.Struct):
    opcode          = btype.uint16_t()
    tag             = btype.uint16_t()
    data_len        = btype.uint16_t()
    rsrv            = btype.uint16_t()
    params          = btype.Array(btype.uint32_t(), 6)
    _EXPECTED_SIZE  = 32


class Response(btype.Struct):
    opcode         = btype.uint16_t()
    tag            = btype.uint16_t()
    data_len       = btype.uint16_t()
    status         = btype.uint16_t()
    params         = btype.Array(btype.uint32_t(), 6)
    _EXPECTED_SIZE = 32


class CommandException(Exception):
    def __init__(self, rsp, rx_data):
        super().__init__(
            'Command exception: %s (%s)' % (Status.rsp_to_status_str(rsp), rsp))
        self.rsp = rsp
        self.rx_data = rx_data


class Device:
    def __init__(self, usb_dev, cmd_ep, rsp_ep, max_cmd_data_len,
                 max_rsp_data_len, git_sha1_index=None,
                 default_configuration=None):
        super().__init__()

        self.usb_dev          = usb_dev
        self.cmd_ep           = cmd_ep
        self.rsp_ep           = rsp_ep
        self.max_cmd_data_len = max_cmd_data_len
        self.max_rsp_data_len = max_rsp_data_len
        self.tag              = random.randint(0, 65535)
        self.last_time_ns     = 0

        try:
            self.serial_num = usb_dev.serial_number
            if git_sha1_index is not None:
                self.git_sha1 = usb.util.get_string(usb_dev, git_sha1_index)
            else:
                self.git_sha1 = None
            self.fw_version = usb_dev.bcdDevice
        except ValueError as e:
            if str(e) == 'The device has no langid':
                raise Exception(
                    'Device has no langid, ensure running as root!') from e

        self.fw_version_str = '%u.%u.%u' % (
                (self.fw_version >> 8) & 0xFF,
                (self.fw_version >> 4) & 0x0F,
                (self.fw_version >> 0) & 0x0F)

        if default_configuration is not None:
            self._set_configuration(default_configuration)
        self._synchronize()

    def _exec_command(self, opcode, params=None, data=b'', timeout_ms=1000):
        if data:
            assert len(data) <= self.max_cmd_data_len
        if not params:
            params = [0, 0, 0, 0, 0, 0]
        elif len(params) < 6:
            params = params + [0]*(6 - len(params))

        tag = self.tag
        self.tag = (self.tag + 1) & 0xFFFF

        cmd = Command(opcode=opcode, tag=tag, data_len=len(data),
                      params=params)
        l = self.usb_dev.write(self.cmd_ep, cmd.pack(), timeout=timeout_ms)
        assert l == 32

        if data:
            l = self.usb_dev.write(self.cmd_ep, data, timeout=timeout_ms)
            assert l == len(data)

        data = self.usb_dev.read(self.rsp_ep, Response._STRUCT.size,
                                 timeout=timeout_ms)
        assert len(data) == Response._STRUCT.size
        rsp = Response.unpack(data)
        assert rsp.opcode == opcode
        assert rsp.tag    == tag

        if rsp.data_len:
            rx_data = bytes(self.usb_dev.read(self.rsp_ep, rsp.data_len,
                                              timeout=timeout_ms))
            assert len(rx_data) == rsp.data_len
        else:
            rx_data = None

        if rsp.status != Status.OK:
            raise CommandException(rsp, rx_data)

        return rsp, rx_data

    def _set_configuration(self, bConfigurationValue):
        cfg = None
        try:
            cfg = self.usb_dev.get_active_configuration()
        except usb.core.USBError as e:
            if e.strerror != 'Configuration not set':
                raise

        if cfg is None or cfg.bConfigurationValue != bConfigurationValue:
            usb.util.dispose_resources(self.usb_dev)
            self.usb_dev.set_configuration(bConfigurationValue)

    def _synchronize(self):
        # Try to abort any existing command.  Depending on the adapter's state:
        #   Waiting to RX CMD: will send a US_ABORTED RSP.
        #   Waiting to RX DATA: will send a US_ABORTED RSP.
        #   Trying to TX RSP: write will time out, read skipped over.
        #   Trying to TX DATA: write will time out, read skipped over.
        #
        # The best way to abort this would be to write a zero-length packet to
        # CMD_EP.  The write should time out and there is no danger of the
        # write completing a data phase.  However, on Windows a zero-length
        # write doesn't cause a timeout, it just silently succeeds even if the
        # target isn't accepting data on that endpoint.  Since we rely on the
        # timeout to detect what state we are in, that breaks the algorithm.
        # So, on Windows we do a 1-byte write instead, which hopefully is safe
        # in almost every case.
        try:
            if platform.system() == 'Windows':
                self.usb_dev.write(self.cmd_ep, b'\x00', timeout=100)
            else:
                self.usb_dev.write(self.cmd_ep, b'', timeout=100)
            data = self.usb_dev.read(self.rsp_ep, 32, timeout=100)
            assert len(data) == 32
            rsp = Response.unpack(data)
            if platform.system() == 'Windows':
                assert rsp.status in (Status.BAD_CMD_LENGTH,
                                      Status.BAD_DATA_LENGTH)
            else:
                assert rsp.status == Status.ABORTED
            return
        except usb.core.USBTimeoutError:
            pass

        # The adapter is either transmitting a RSP or a DATA sequence.  We can't
        # be sure which so we need to do two reads.  The first read will always
        # succeed and the second read will time out if there was no DATA phase
        # associated with the RSP.
        self.usb_dev.read(self.rsp_ep, self.max_rsp_data_len, timeout=100)
        try:
            self.usb_dev.read(self.rsp_ep, self.max_rsp_data_len, timeout=100)
        except usb.core.USBTimeoutError:
            pass

    def time_ns_increasing(self):
        '''
        Returns a time value in nanoseconds that is guaranteed to increase
        after every single call.  This function is not thread-safe.
        '''
        self.last_time_ns = t = max(time.time_ns(), self.last_time_ns + 1)
        return t



# Copyright (c) 2022-2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import time
import struct

import xtalx.modbus_adapter

from .xti import Measurement
from .cal_page import CalPage


PARITY_DICT = {
    'N' : 0,
    'E' : 2,
    'O' : 3,
}


class XHTISM:
    '''
    This is a driver for the high-temperature XHTIS sensor built with Modbus
    firmware.
    '''
    def __init__(self, bus, slave_addr=0x80):
        self.bus          = bus
        self.slave_addr   = slave_addr
        self._halt_yield  = True
        self.last_time_ns = 0

        (self.serial_num,
         self.fw_version_str,
         self.fw_version,
         self.git_sha1) = self._read_ids()

        self.cal_page = self.read_valid_calibration_page()

        self.report_id = None
        self.poly_psi  = None
        self.poly_temp = None
        if self.cal_page is not None:
            self.report_id = self.cal_page.get_report_id()
            self.poly_psi, self.poly_temp = self.cal_page.get_polynomials()

    def __str__(self):
        return 'XHTISM(%s)' % self.serial_num

    def _read_ids(self):
        objs = self.bus.read_device_identification(self.slave_addr, 0x01, 0x00)

        assert objs[0].value == b'Phase'
        assert objs[1].value.startswith(b'XHTIS-')

        fw_version_str = objs[2].value.decode()
        parts          = fw_version_str.split('.')
        fw_version = ((int(parts[0]) << 8) |
                      (int(parts[1]) << 4) |
                      (int(parts[2]) << 0))

        git_sha1 = self.bus.read_holding_registers_binary(self.slave_addr,
                                                          0x1000, 23)

        return (objs[1].value.decode(), fw_version_str, fw_version,
                git_sha1.decode().strip())

    def set_comm_params(self, baud_rate, slave_addr, parity):
        '''
        Set the sensor's baud rate and Modbus slave address.  The baud rate
        must be a multiple of 4800.  As per the Modbus specification, even
        parity is always used.  These values are saved on the sensor in
        nonvolatile storage.

        After the command executes on the sensor, the Modbus reply is returned
        using the original slave address and baud rate.  After the final byte
        of the reply is transmitted, the sensor resets itself and starts back
        up using the new slave address and baud rate.  Due to this reset, there
        may be an additional delay before the sensor responds to the next
        command at the new address and baud rate.

        In the event of an error storing the values to nonvolatile storage,
        the sensor returns an error code and does not reset or switch to the
        new parameters.
        '''
        assert baud_rate % 4800 == 0
        baud_rate = baud_rate // 4800
        assert baud_rate < 64
        assert slave_addr < 256
        self.bus.write_holding_registers_binary(self.slave_addr, 0x1001,
                                                bytes([baud_rate, slave_addr,
                                                       PARITY_DICT[parity],
                                                       0x00]))

    def get_coefficients(self):
        '''
        Returns the T and P IIR filter coefficients currently in-use by the
        sensor.  These values are saved on the sensor in nonvolatile storage.
        '''
        rsp = self.bus.read_holding_registers_binary(self.slave_addr, 0x1002, 1)
        t_c = rsp[0]
        p_c = rsp[1]
        return t_c, p_c

    def set_coefficients(self, t_c, p_c):
        '''
        Set the T and P IIR filter coefficients used by the sensor.  These
        values are saved on the sensor in nonvolatile storage.  The new values
        take effect immediately; no sensor reset required.
        '''
        assert t_c < 32
        assert p_c < 32
        self.bus.write_holding_registers_binary(self.slave_addr, 0x1002,
                                                bytes([t_c, p_c]))

    def read_calibration_pages_raw(self):
        '''
        Returns the raw data bytes for the single calibration page stored in
        flash.
        '''
        data = b''
        for i in range(CalPage.get_short_size() // 8):
            address = 0x2000 + i*4
            data += self.bus.read_holding_registers_binary(self.slave_addr,
                                                           address, 4)
        pad = b'\xff' * (CalPage._EXPECTED_SIZE - len(data))
        return (data + pad,)

    def read_calibration_pages(self):
        '''
        Returns a CalPage struct for the single calibration page in sensor
        flash, even if the page is missing or corrupt.
        '''
        (cp_data,) = self.read_calibration_pages_raw()
        cp = CalPage.unpack(cp_data)
        return (cp,)

    def read_valid_calibration_page(self):
        '''
        Returns CalPage struct from the sensor flash.  Returns None if the
        calibration is not present or corrupted.
        '''
        (cp,) = self.read_calibration_pages()
        return cp if cp.is_valid() else None

    def read_measurement(self):
        rsp = self.bus.read_holding_registers_binary(self.slave_addr, 0, 8)

        ft, fp = struct.unpack('<dd', rsp)

        psi = None
        if self.poly_psi is not None and ft and fp:
            psi = self.poly_psi(fp, ft)

        temp_c = None
        if self.poly_temp is not None and ft:
            temp_c = self.poly_temp(ft)

        m = Measurement(self, None, psi, temp_c, fp, ft, None, None,
                        None, None, None, None, None, None, None)
        if isinstance(self.bus, xtalx.modbus_adapter.MBA):
            m._current_amps = self.bus.measure_current()

        return m

    def yield_measurements(self, poll_interval_sec=0.1, **_kwargs):
        self._halt_yield = False
        while not self._halt_yield:
            yield self.read_measurement()
            time.sleep(poll_interval_sec)

    def halt_yield(self):
        self._halt_yield = True

    def time_ns_increasing(self):
        '''
        Returns a time value in nanoseconds that is guaranteed to increase
        after every single call.  This function is not thread-safe.
        '''
        self.last_time_ns = t = max(time.time_ns(), self.last_time_ns + 1)
        return t



# Copyright (c) 2025 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import btype

from .xti import Measurement
from .exception import XtalXException


class FrequencyResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    pressure_hz    = btype.float64_t()
    temperature_hz = btype.float64_t()
    _EXPECTED_SIZE = 17


class ConversionResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    pressure_psi   = btype.float64_t()
    temperature_c  = btype.float64_t()
    _EXPECTED_SIZE = 17


class FixedResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    pressure_psi   = btype.int32_t()
    temperature_c  = btype.int32_t()
    _EXPECTED_SIZE = 9


class FullResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    pressure_psi   = btype.float64_t()
    temperature_c  = btype.float64_t()
    pressure_hz    = btype.float64_t()
    temperature_hz = btype.float64_t()
    _EXPECTED_SIZE = 33


class DeadFirmwareException(XtalXException):
    '''
    While probing, the firmware repeatedly returned '????'.
    '''
    def __str__(self):
        return 'DeadFirmwareException()'


class Comms:
    '''
    Communication protocol for sensor firmware 0.9.1 or below.
    '''
    def __init__(self, xhtiss):
        self.xhtiss = xhtiss

    def _set_nvstore(self, addr, data):
        assert len(data) == 4
        cmd = bytes([0x1E, 0x00, (addr & 0xFF), ((addr >> 8) & 0xFF),
                     data[0], data[1], data[2], data[3]])
        self.xhtiss.bus.transact(cmd)

    def _get_nvstore(self, addr, size):
        cmd = bytes([0x2A, 0x00, (addr & 0xFF), ((addr >> 8) & 0xFF), 0x00])
        cmd += bytes(size)
        data = self.xhtiss.bus.transact(cmd)
        return data[5:]

    def exec_cmd(self, cmd, rsp_len):
        cmd_bytes = bytes([cmd, 0x00]) + b'\x00'*rsp_len
        rsp = self.xhtiss.bus.transact(cmd_bytes)
        if rsp == b'?'*len(rsp):
            raise DeadFirmwareException()
        return rsp[2:]

    def read_frequencies(self):
        data = self.exec_cmd(0x19, FrequencyResponse._STRUCT.size)
        return FrequencyResponse.unpack(data)

    def read_conversion(self):
        data = self.exec_cmd(0x07, ConversionResponse._STRUCT.size)
        return ConversionResponse.unpack(data)

    def read_fixed(self):
        data = self.exec_cmd(0x33, FixedResponse._STRUCT.size)
        return FixedResponse.unpack(data)

    def read_full(self):
        data = self.exec_cmd(0x2D, FullResponse._STRUCT.size)
        return FullResponse.unpack(data)

    def read_measurement(self):
        rsp = self.read_full()

        m = Measurement(self.xhtiss, None, rsp.pressure_psi, rsp.temperature_c,
                        rsp.pressure_hz, rsp.temperature_hz, None, None, None,
                        None, None, None, None, None, None)
        m._age_ms = rsp.age_ms
        return m



#!/usr/bin/env python3
# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import threading
import argparse
import logging
import time

from xtalx.tools.config import Config
from xtalx.tools.influxdb import InfluxDBPushQueue
import xtalx.p_sensor
import xtalx.modbus_adapter
import xtalx.tools.modbus.serial


# Verbosity
LOG_LEVEL = logging.INFO


def sensor_thread(xhtism, idb):
    # Monitor the sensor.
    logging.info('%s: Monitoring...', xhtism.serial_num)
    for m in xhtism.yield_measurements():
        if idb:
            point = m.to_influx_point()
            if point['fields']:
                idb.append(point)

        logging.info('%s: tf %.6f pf %.6f ',
                     xhtism.serial_num, m.temp_freq, m.pressure_freq)


def main(rv):
    # Read the configuration file.
    if rv.config:
        logging.info('Reading configuration...')
        with open(rv.config, encoding='utf8') as f:
            c = Config(f.readlines(), ['influx_host', 'influx_user',
                                       'influx_password', 'influx_database'])

        # Open a connection to InfluxDB.
        logging.info('Connecting to InfluxDB...')
        idb = InfluxDBPushQueue(c.influx_host, 8086, c.influx_user,
                                c.influx_password, database=c.influx_database,
                                ssl=True, verify_ssl=True,
                                timeout=100, throttle_secs=10)
    else:
        idb = None

    # Make the bus.
    if rv.intf:
        bus = xtalx.tools.modbus.serial.Bus(rv.intf, rv.baud_rate)
    else:
        dev = xtalx.modbus_adapter.find_one_mba(
                serial_number=rv.modbus_adapter_serial_number)
        if dev is None:
            raise Exception('No adapter found.')
        bus = xtalx.modbus_adapter.make_mba(dev, baud_rate=rv.baud_rate)
        bus.set_vext(True)
        time.sleep(0.1)

    # Spawn a thread for each sensor on the bus.
    xhtisms = []
    threads = []
    addrs = rv.addr
    if addrs is None:
        addrs = ['0x80']
    for addr in addrs:
        xhtism = xtalx.p_sensor.XHTISM(bus, int(addr, 0))
        logging.info('%s: Found sensor with firmware version %s, git SHA1 %s',
                     xhtism.serial_num, xhtism.fw_version_str, xhtism.git_sha1)
        t_c, p_c = xhtism.get_coefficients()
        logging.info('%s: T Coefficient: %u', xhtism.serial_num, t_c)
        logging.info('%s: P Coefficient: %u', xhtism.serial_num, p_c)
        xhtisms.append(xhtism)

        t = threading.Thread(target=sensor_thread, args=(xhtism, idb))
        t.start()
        threads.append(t)

    # Run forever.
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print()

    # Halt and join all threads for a clean exit.
    for xhtism in xhtisms:
        xhtism.halt_yield()
    for t in threads:
        t.join()


def _main():
    logging.basicConfig(format='\033[1m[%(asctime)s.%(msecs)03d]\033[0m '
                        '%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logging.getLogger().setLevel(LOG_LEVEL)

    parser = argparse.ArgumentParser()
    parser.add_argument('--config', '-c')
    parser.add_argument('--intf', '-i')
    parser.add_argument('--baud-rate', '-b', default=115200, type=int)
    parser.add_argument('--addr', '-a', action='append', default=None)
    parser.add_argument('--modbus-adapter-serial-number', '-s')

    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



#!/usr/bin/env python3
# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import logging
import time

import xtalx.p_sensor
import xtalx.modbus_adapter
import xtalx.tools.modbus.serial


# Verbosity
LOG_LEVEL = logging.INFO


def make_sensor(args):
    if args.intf:
        bus = xtalx.tools.modbus.serial.Bus(args.intf, args.baud_rate,
                                            parity=args.parity)
        return xtalx.p_sensor.XHTISM(bus, int(args.addr, 0))

    dev = xtalx.modbus_adapter.find_one_mba(
            serial_number=args.modbus_adapter_serial_number)
    if dev is not None:
        bus = xtalx.modbus_adapter.make_mba(dev, baud_rate=args.baud_rate,
                                            parity=args.parity)
        bus.set_vext(True)
        time.sleep(0.1)
        return xtalx.p_sensor.XHTISM(bus, int(args.addr, 0))

    raise Exception('No matching devices.')


def main(args):
    # Make the sensor.
    xhtism = make_sensor(args)
    logging.info('%s: Found sensor with firmware version %s, git SHA1 %s',
                 xhtism.serial_num, xhtism.fw_version_str, xhtism.git_sha1)
    t_c, p_c = xhtism.get_coefficients()
    logging.info('%s: T Coefficient: %u', xhtism.serial_num, t_c)
    logging.info('%s: P Coefficient: %u', xhtism.serial_num, p_c)

    # If we need to update the T/P coefficients, do so now.
    if args.set_t_coefficient or args.set_p_coefficient:
        if args.set_t_coefficient:
            t_c = int(args.set_t_coefficient, 0)
        if args.set_p_coefficient:
            p_c = int(args.set_p_coefficient, 0)
        logging.info('%s: Updating coefficients...', xhtism.serial_num)
        xhtism.set_coefficients(t_c, p_c)

    # If we need to change the comms params, now is the time.
    if args.set_addr or args.set_baud_rate or args.set_parity:
        logging.info('%s: Updating comm params...', xhtism.serial_num)
        if args.set_baud_rate:
            new_baud_rate = args.set_baud_rate
        else:
            new_baud_rate = args.baud_rate
        if args.set_parity:
            new_parity = args.set_parity
        else:
            new_parity = args.parity
        if args.set_addr:
            addr = int(args.set_addr, 0)
        else:
            addr = xhtism.slave_addr
        xhtism.set_comm_params(new_baud_rate, addr, new_parity)


def _main():
    logging.basicConfig(format='\033[1m[%(asctime)s.%(msecs)03d]\033[0m '
                        '%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logging.getLogger().setLevel(LOG_LEVEL)

    parser = argparse.ArgumentParser()
    parser.add_argument('--intf', '-i')
    parser.add_argument('--baud-rate', '-b', default=115200, type=int)
    parser.add_argument('--parity', default='E', choices=['E', 'O', 'N'])
    parser.add_argument('--addr', '-a', default='0x80')
    parser.add_argument('--set-addr')
    parser.add_argument('--set-baud-rate', type=int)
    parser.add_argument('--set-parity', choices=['E', 'O', 'N'])
    parser.add_argument('--set-t-coefficient')
    parser.add_argument('--set-p-coefficient')
    parser.add_argument('--modbus-adapter-serial-number', '-s')

    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2025 Phase Advanced Sensor Systems, Inc.
from enum import IntEnum

import xtalx.tools.modbus
import xtalx.usbcmd


class Feature(IntEnum):
    CURRENT_MEASUREMENT     = (1 << 0)


class Status(IntEnum):
    XACT_FAILED     = 7
    XACT_FAILED_2   = 100

    @staticmethod
    def rsp_to_status_str(rsp):
        try:
            return '%s' % Status(rsp.status)
        except ValueError:
            return xtalx.usbcmd.Status.rsp_to_status_str(rsp)


class Opcode(IntEnum):
    SET_VEXT                 = 0x0001
    SET_BAUD_RATE            = 0x0002
    XACT                     = 0x0003
    MEASURE_CURRENT          = 0x0005
    JUMP_BOOTLOADER          = 0x7B42


CMD_EP = 0x01
RSP_EP = 0x82


PARITY_DICT = {
    'N' : 0x4E4F4E45,
    'E' : 0x4556454E,
    'O' : 0x4F444420,
}


class MBA(xtalx.tools.modbus.Bus,
          xtalx.usbcmd.Device):

    def __init__(self, usb_dev, baud_rate=115200, parity='E'):
        xtalx.tools.modbus.Bus.__init__(self)
        xtalx.usbcmd.Device.__init__(self, usb_dev, CMD_EP, RSP_EP, 256, 256,
                                     git_sha1_index=6,
                                     default_configuration=0x80)

        if self.fw_version < 0x093:
            self.xact_failed_code = Status.XACT_FAILED
        else:
            self.xact_failed_code = Status.XACT_FAILED_2

        self.features = 0
        if self.serial_num.startswith('MBA-2'):
            if self.fw_version >= 0x092:
                self.features |= Feature.CURRENT_MEASUREMENT

        self.set_comm_params(baud_rate, parity)

    def __str__(self):
        return 'MBA(%s)' % self.serial_num

    def set_vext(self, enabled):
        return self._exec_command(Opcode.SET_VEXT, [enabled])

    def set_comm_params(self, baud_rate, parity=None):
        params = [baud_rate, 0, 0, PARITY_DICT.get(parity, 0)]
        return self._exec_command(Opcode.SET_BAUD_RATE, params)

    def measure_current(self):
        if (self.features & Feature.CURRENT_MEASUREMENT) == 0:
            return 0.0

        rsp, _ = self._exec_command(Opcode.MEASURE_CURRENT)
        adc_iir = (((rsp.params[1] << 32) & 0xFFFFFFFF00000000) |
                   ((rsp.params[0] <<  0) & 0x00000000FFFFFFFF))
        return (3.3 / (4096 * 1000 * 0.033 * 2**52)) * adc_iir

    def enter_dfu_mode(self):
        return self._exec_command(Opcode.JUMP_BOOTLOADER, [0xA47B39FE])

    def transact(self, addr, data, response_time_ms, nbytes=None):
        try:
            _, rx_data = self._exec_command(Opcode.XACT,
                                            [addr, response_time_ms], data,
                                            timeout_ms=(response_time_ms + 100))
        except xtalx.usbcmd.CommandException as e:
            if e.rsp.status != self.xact_failed_code:
                raise

            if e.rsp.params[0] == 0:
                raise xtalx.tools.modbus.ResponseOverflowException(b'')
            if e.rsp.params[0] == 1:
                raise xtalx.tools.modbus.ResponseUnderflowException(b'')
            if e.rsp.params[0] == 2:
                raise xtalx.tools.modbus.BadCRCException(b'', None)
            if e.rsp.params[0] == 3:
                raise xtalx.tools.modbus.BadAddressException(b'')
            if e.rsp.params[0] == 4:
                raise xtalx.tools.modbus.BabbleException(b'')
            if e.rsp.params[0] == 5:
                raise xtalx.tools.modbus.ResponseInterruptedException(b'')
            if e.rsp.params[0] == 6:
                raise xtalx.tools.modbus.ResponseFramingErrorException(b'')
            if e.rsp.params[0] == 7:
                raise xtalx.tools.modbus.ResponseTimeoutException(b'')

            raise

        # Check the response function code.
        if rx_data[1] & 0x7F != data[0]:
            raise xtalx.tools.modbus.BadFunctionException(rx_data)
        if rx_data[1] & 0x80:
            raise xtalx.tools.modbus.ExceptionResponseException(rx_data,
                                                                rx_data[2])

        return rx_data



# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import threading


class ModbusException(Exception):
    pass


class ResponseException(ModbusException):
    def __init__(self, rx_packet, **kwargs):
        super().__init__(rx_packet.hex(), **kwargs)
        self.rx_packet = rx_packet


class ExceptionResponseException(ResponseException):
    def __init__(self, rx_packet, exception_number, **kwargs):
        super().__init__(rx_packet, **kwargs)
        self.exception_number = exception_number


class ResponseTimeoutException(ResponseException):
    pass


class BadCRCException(ResponseException):
    def __init__(self, rx_packet, expected_crc, **kwargs):
        super().__init__(rx_packet, **kwargs)
        self.expected_crc = expected_crc


class BadAddressException(ResponseException):
    pass


class BadFunctionException(ResponseException):
    pass


class BabbleException(ResponseException):
    # Target replied too quickly, before t3.5 expired.
    pass


class ResponseOverflowException(ResponseException):
    pass


class ResponseUnderflowException(ResponseException):
    pass


class ResponseInterruptedException(ResponseException):
    # We received some characters, then t1.5 expired, then we received more
    # characters.
    pass


class ResponseFramingErrorException(ResponseException):
    pass


class ResponseSyntaxException(ResponseException):
    pass


class DeviceIDObject:
    def __init__(self, object_id, value):
        self.object_id = object_id
        self.value     = value


class Bus:
    def __init__(self):
        self.lock = threading.Lock()

    def transact(self, addr, data, response_time_ms, nbytes=None):
        '''
        Perform a command/response transaction on the bus, initiated by sending
        the specified data bytes to the specified address.  The response_time_ms
        timeout is how long we are willing to wait for the first byte of the
        response.  This method returns the response data as a bytes() object.

        If response_time_ms is 0, then no response is expected (perhaps this was
        a broadcast write), we don't try to read one, and this method will
        return None instead.

        The nbytes parameter is an optional hint to the implementation on how
        many bytes to expect in a response.  It is useful for the serial
        implementation which can't meet the strict Modbus timing requirements
        to delimit PDUs.
        '''
        raise NotImplementedError

    def read_device_identification(self, slave_addr, read_code, object_id,
                                   response_time_ms=100):
        '''
        Performs a Read Device Identification request with the specified Read
        Device ID code and Object Id.

        Returns a list of DeviceIDObject objects, which have object_id and
        value fields corresponding to the same fields in the Modbus spec.
        '''
        with self.lock:
            rsp = self.transact(slave_addr, bytes([0x2B, 0x0E, read_code,
                                                   object_id]),
                                response_time_ms)

        nobjs  = rsp[7]
        objs   = []
        offset = 8
        for _ in range(nobjs):
            object_id = rsp[offset]
            size      = rsp[offset + 1]
            value     = rsp[offset + 2: offset + 2 + size]
            offset   += 2 + size
            objs.append(DeviceIDObject(object_id, value))

        return objs

    def read_holding_registers_binary(self, slave_addr, address, nregs,
                                      response_time_ms=100):
        '''
        Reads nregs 16-bit registers from the target device slave_addr starting
        with register address.  The return value is the raw binary data in the
        exact order it was streamed from the serial bus.  No splitting into
        16-bit values or byte-swapping of any sort is done; this returns the
        raw data.
        '''
        nbytes = 2 + nregs * 2
        with self.lock:
            rsp = self.transact(slave_addr, bytes([0x03,
                                                   (address >> 8) & 0xFF,
                                                   address & 0xFF,
                                                   (nregs >> 8) & 0xFF,
                                                   nregs & 0xFF]),
                                response_time_ms, nbytes=nbytes)
        if rsp[2] != 2 * nregs:
            raise ResponseSyntaxException(rsp)

        return rsp[3:-2]

    def write_holding_registers_binary(self, slave_addr, address, data,
                                       response_time_ms=100):
        '''
        Writes data to the target device slave_addr at register address.  The
        data is streamed out in order and no byte-swapping is performed.  Try
        to keep writes short so that there is no danger of a write being broken
        up in a kernel or adapter buffer.
        '''
        assert len(data) % 2 == 0
        assert 1 <= len(data) <= 123
        nregs = len(data) // 2
        with self.lock:
            rsp = self.transact(slave_addr, bytes([0x10,
                                                   (address >> 8) & 0xFF,
                                                   address & 0xFF,
                                                   (nregs >> 8) & 0xFF,
                                                   nregs & 0xFF,
                                                   len(data)]) + data,
                                response_time_ms, nbytes=5)
        if ((rsp[2] << 8) | rsp[3]) != address:
            raise ResponseSyntaxException(rsp)
        if ((rsp[4] << 8) | rsp[5]) != nregs:
            raise ResponseSyntaxException(rsp)



#!/usr/bin/env python3
# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import logging
import time

import xtalx.p_sensor
import xtalx.spi_adapter


# Verbosity
LOG_LEVEL = logging.INFO


def make_sensor(args):
    dev = xtalx.spi_adapter.find_one_spia()
    if dev is not None:
        bus = xtalx.spi_adapter.make_spia(dev)
        bus.set_vext(True)
        time.sleep(0.2)
        x = xtalx.p_sensor.XHTISS(bus)
        if args.serial_number is None or x.serial_num == args.serial_number:
            x.poll_interval_sec = 0.2
            return x

    raise Exception('No matching devices.')


def main(args):
    # Make the sensor.
    xhtiss = make_sensor(args)
    logging.info('%s: Found sensor with firmware version %s, git SHA1 %s',
                 xhtiss.serial_num, xhtiss.fw_version_str, xhtiss.git_sha1)
    t_c, p_c, sample_ms = xhtiss.get_flash_params()
    logging.info('%s: T Coefficient: %u', xhtiss.serial_num, t_c)
    logging.info('%s: P Coefficient: %u', xhtiss.serial_num, p_c)
    logging.info('%s:   Sample Rate: %u ms', xhtiss.serial_num, sample_ms)

    # If we need to update the T/P coefficients, do so now.
    if args.set_t_coefficient or args.set_p_coefficient:
        if args.set_t_coefficient:
            t_c = int(args.set_t_coefficient, 0)
        if args.set_p_coefficient:
            p_c = int(args.set_p_coefficient, 0)
        if args.set_sample_ms:
            sample_ms = int(args.set_sample_ms, 0)
        logging.info('%s: Updating coefficients...', xhtiss.serial_num)
        xhtiss.set_flash_params(t_c, p_c, sample_ms)


def _main():
    logging.basicConfig(format='\033[1m[%(asctime)s.%(msecs)03d]\033[0m '
                        '%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logging.getLogger().setLevel(LOG_LEVEL)

    parser = argparse.ArgumentParser()
    parser.add_argument('--intf', '-i')
    parser.add_argument('--set-t-coefficient')
    parser.add_argument('--set-p-coefficient')
    parser.add_argument('--set-sample-ms')
    parser.add_argument('--serial-number', '-s')

    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2022-2024 by Phase Advanced Sensor Systems, Inc.
import struct


def hex_to_double(v):
    '''
    Converts a 64-bit integer representation of a double-precision floating
    point number into its actual floating-point value.  I.e., the Python
    equivalent of this C union:

        union
        {
            uint64_t    u;
            double      d;
        };
    '''
    return struct.unpack('<d', struct.pack('<Q', v))[0]


def hex_s_to_double(v):
    '''
    Converts a 64-bit value encoded as a hex string into a floating-point
    double value.
    '''
    return hex_to_double(int(v, 16))



# Copyright (c) 2020-2023 by Phase Advanced Sensor Systems Corp.
import xtalx.tools.usb

from .xhti import XHTI
from .xhtism import XHTISM
from .xhtiss import XHTISS
from .xmhti import XMHTI
from .xti import XTI


def find_xti(**kwargs):
    return xtalx.tools.usb.find(idVendor=0x0483, idProduct=0xA34E,
                                product='XtalX', find_all=True, **kwargs)


def find_one_xti(**kwargs):
    return xtalx.tools.usb.find_one(idVendor=0x0483, idProduct=0xA34E,
                                    product='XtalX', find_all=True, **kwargs)


def make_xti(usb_dev, **kwargs):
    if usb_dev.product == 'XtalX':
        return XTI(usb_dev, **kwargs)

    raise Exception('Unrecognized product string: %s' % usb_dev.product)


def find_xmhti(**kwargs):
    return xtalx.tools.usb.find(idVendor=0x0483, idProduct=0xA34E,
                                bDeviceClass=0xFF, bDeviceSubClass=0x05,
                                find_all=True, **kwargs)


def find_one_xmhti(**kwargs):
    return xtalx.tools.usb.find_one(idVendor=0x0483, idProduct=0xA34E,
                                    bDeviceClass=0xFF, bDeviceSubClass=0x05,
                                    find_all=True, **kwargs)


def make_xmhti(usb_dev, **kwargs):
    if usb_dev.bDeviceClass == 0xFF and usb_dev.bDeviceSubClass == 0x05:
        return XMHTI(usb_dev, **kwargs)

    raise Exception('Unrecognized device: %s' % usb_dev)


def make(usb_dev, **kwargs):
    if usb_dev.bDeviceClass == 0xFF and usb_dev.bDeviceSubClass == 0x05:
        return XMHTI(usb_dev, **kwargs)
    if usb_dev.product == 'XtalX':
        return XTI(usb_dev, **kwargs)

    raise Exception('Unrecognized product string: %s' % usb_dev.product)


__all__ = ['find_xti',
           'find_one_xti',
           'make_xti',
           'find_xmhti',
           'find_one_xmhti',
           'make_xmhti',
           'make',
           'XTI',
           'XHTI',
           'XHTISM',
           'XHTISS',
           'XMHTI',
           ]



# Copyright (c) 2025 by Phase Advanced Sensor Systems Corp.
import xtalx.tools.usb

from .modbus_adapter import MBA


def find_mba(**kwargs):
    return xtalx.tools.usb.find(idVendor=0x0483, idProduct=0xA34E,
                                bDeviceClass=0xFF, bDeviceSubClass=0x0C,
                                find_all=True, **kwargs)


def find_one_mba(**kwargs):
    return xtalx.tools.usb.find_one(idVendor=0x0483, idProduct=0xA34E,
                                    bDeviceClass=0xFF, bDeviceSubClass=0x0C,
                                    find_all=True, **kwargs)


def make_mba(usb_dev, **kwargs):
    if usb_dev.bDeviceClass == 0xFF and usb_dev.bDeviceSubClass == 0x0C:
        return MBA(usb_dev, **kwargs)

    raise Exception('Unrecognized device: %s' % usb_dev)


__all__ = ['find_mba',
           'find_one_mba',
           'make_mba',
           'MBA',
           ]



#!/usr/bin/env python3
# Copyright (c) 2025 by Phase Advanced Sensor Systems Corp.
import argparse
import threading
import time

import xtalx.p_sensor
import xtalx.modbus_adapter
import xtalx.spi_adapter
import xtalx.tools.modbus.serial


def measure_thread(x):
    for m in x.yield_measurements(do_reset=False):
        print(m.tostring(verbose=True))


def make_sensor(args):
    if args.intf:
        bus = xtalx.tools.modbus.serial.Bus(args.intf, args.baud_rate)
        x = xtalx.p_sensor.XHTISM(bus, int(args.modbus_addr, 0))
        if args.serial_number is None or x.serial_num == args.serial_number:
            return x

    dev = xtalx.p_sensor.find_one_xti(serial_number=args.serial_number)
    if dev is not None:
        return xtalx.p_sensor.make(dev)

    dev = xtalx.modbus_adapter.find_one_mba()
    if dev is not None:
        bus = xtalx.modbus_adapter.make_mba(dev, baud_rate=args.baud_rate)
        bus.set_vext(True)
        time.sleep(0.2)
        x = xtalx.p_sensor.XHTISM(bus, int(args.modbus_addr, 0))
        if args.serial_number is None or x.serial_num == args.serial_number:
            return x

    dev = xtalx.spi_adapter.find_one_spia()
    if dev is not None:
        bus = xtalx.spi_adapter.make_spia(dev)
        bus.set_vext(True)
        time.sleep(0.2)
        x = xtalx.p_sensor.XHTISS(bus)
        if args.serial_number is None or x.serial_num == args.serial_number:
            x.poll_interval_sec = 0.2
            return x

    raise Exception('No matching devices.')


def main(args):
    x = make_sensor(args)
    print('Found sensor: %s' % x.serial_num)
    print('  FW Version: %s' % x.fw_version_str)
    print('    Git SHA1: %s' % x.git_sha1)

    mt = threading.Thread(target=measure_thread, args=(x,))
    mt.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print()
    finally:
        x.halt_yield()
        mt.join()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--intf', '-i')
    parser.add_argument('--baud-rate', type=int, default=115200)
    parser.add_argument('--modbus-addr', '-m', default='0x80')
    parser.add_argument('--serial-number', '-s')
    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



#!/usr/bin/env python3
# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import argparse
import time
import sys

import xtalx.modbus_adapter
import xtalx.usbcmd
from xtalx.tools.iter import prange


# Baud rates to probe.
DEFAULT_BAUD_RATES = [115200, 57600]
EXCLUDE_BAUD_RATES = []
BAUD_RATES = DEFAULT_BAUD_RATES + [
    4800*i for i in range(1, 256)
    if 4800*i not in (DEFAULT_BAUD_RATES + EXCLUDE_BAUD_RATES)
]


def main(args):
    # Find the XtalX Modbus adapter.
    dev = xtalx.modbus_adapter.find_one_mba(
            serial_number=args.modbus_adapter_serial_number)
    if dev is None:
        print('No XtalX Modbus adapter found.')
        return

    # Make the bus.
    bus = xtalx.modbus_adapter.make_mba(dev, baud_rate=BAUD_RATES[0],
                                        parity=args.parity)

    # Search.
    print('Searching...')
    found_devices = {}
    try:
        rem_sensors = args.num_sensors
        while rem_sensors > 0:
            while BAUD_RATES:
                if rem_sensors == 0:
                    break
                baud_rate = BAUD_RATES.pop(0)
                print('Trying %u baud.' % baud_rate)
                bus.set_comm_params(baud_rate, args.parity)
                bus.set_vext(False)
                time.sleep(1)
                bus.set_vext(True)
                time.sleep(1)

                for addr in prange(1, 248, clear_line=True):
                    try:
                        objs = bus.read_device_identification(
                            addr, 0x01, 0x00,
                            response_time_ms=args.response_timeout_ms)
                    except xtalx.tools.modbus.ResponseTimeoutException:
                        continue
                    except xtalx.tools.modbus.ModbusException as e:
                        sys.stdout.write('\r**** Device 0x%02X error: '
                                         '%s\x1B[0K\r\n'
                                         % (addr, str(type(e)) + str(e)))
                        continue
                    except xtalx.usbcmd.CommandException as e:
                        sys.stdout.write('\r**** Device 0x%02X error: '
                                         '%s\x1B[0K\r\n'
                                         % (addr, str(type(e)) + str(e)))
                        continue

                    sys.stdout.write('\r**** Device 0x%02X found: %s '
                                     '%s\x1B[0K\r\n'
                                     % (addr, objs[0].value, objs[1].value))
                    found_devices[addr] = (baud_rate, objs)

                    rem_sensors -= 1
                    if rem_sensors == 0:
                        break
    except KeyboardInterrupt:
        print()

    print('Device summary')
    print('==============')
    if found_devices:
        for k, v in found_devices.items():
            baud_rate, objs = v
            print('0x%02X @ %u baud: %s %s'
                  % (k, baud_rate, objs[0].value, objs[1].value))
    else:
        print('No devices found.')


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--modbus-adapter-serial-number', '-s')
    parser.add_argument('--parity', default='E', choices=['E', 'O', 'N'])
    parser.add_argument('--num-sensors', '-n', type=int, default=1)
    parser.add_argument('--response-timeout-ms', '-t', type=int, default=100)

    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



#!/usr/bin/env python3
# Copyright (c) 2020-2023 by Phase Advanced Sensor Systems Corp.
import argparse

import xtalx.p_sensor


VERBOSE = False


def xtalx_cb(m):
    print(m.tostring(VERBOSE))


def main(args):
    global VERBOSE
    VERBOSE = args.verbose

    d = xtalx.p_sensor.find_one_xti(serial_number=args.serial_number)
    x = xtalx.p_sensor.make_xti(d)
    x.read_measurements(xtalx_cb)
    try:
        x.join_read()
    except KeyboardInterrupt:
        x.halt_read()
        raise


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--serial-number', '-s')
    parser.add_argument('--verbose', '-v', action='store_true')
    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2022 by Phase Advanced Sensor Systems.
# All rights reserved.
import argparse
import time
import os

import xtalx.p_sensor
from . import xmhti_qspi_decode


def measure_continuous(xmhti):
    print('Starting continuous mode...')
    xmhti.start_continuous_mode()
    while True:
        ms = xmhti.read_measurements()
        for m in ms:
            print('I:0x%08X  T:%10.3f  P:%9.3f  VDD:%5.3f  '
                  'VBUS/2:%5.3f' % (m.seq, m.ft, m.fp, m.vdd, m.vbus))
        time.sleep(9.9)


def main(args):
    d = xtalx.p_sensor.find_one_xmhti(serial_number=args.serial_number)
    xmhti = xtalx.p_sensor.make_xmhti(d)

    # Get boot status.
    bs = xmhti._boot_status
    ac = xmhti.get_autonomous_config()
    print(' Serial Num: %s' % xmhti.serial_num)
    print(' FW Version: %s' % xmhti.fw_version_str)
    print('       SHA1: %s' % xmhti.git_sha1)
    print('POST Status: 0x%08X' % bs.post_result)
    print('   PLL Freq: %s MHz' % (bs.pll_freq / 1000000))
    print('   HSE Freq: %s MHz' % (bs.hse_freq / 1000000))
    print('    LS Freq: %.3f KHz (%u ticks)' % (bs.ls_freq / 1000, bs.ls_ticks))
    if xmhti.cal_page is not None:
        print('OSC Startup: %u ms' % xmhti.cal_page.osc_startup_time_ms)
    print('Auto Config: %u secs, flags 0x%02X' % (ac.interval_secs, ac.flags))

    # Issue a Get QSPI Flash Info command.
    qi = xmhti.get_qspi_flash_info()
    print('QSPI Dev ID: 0x%02X' % qi.dev_id)
    print('QSPI Man ID: 0x%02X' % qi.m_id)
    print('  QSPI D_ID: 0x%02X' % qi.d_id)
    print('   QSPI Pos: %u / %u' % (qi.write_index, qi.nslots))
    print('  QSPI Size: %s MB'  % (qi.capacity / (1024 * 1024)))

    # Issue a Get Voltages command.
    vs = xmhti.get_voltages()
    print('        VDD: %.3f V' % (vs.vdd_mv / 1000))
    print('       VUSB: %.3f V' % (vs.vbus_mv * 2 / 1000))

    # If requested, perform a chip erase.
    if args.chip_erase_start:
        print('Sending Chip Erase command...')
        xmhti.start_erase_qspi_flash()

    # If requested, wait for a chip erase to complete.
    if args.chip_erase_wait_done:
        print('Waiting for Chip Erase to complete...')
        t0 = time.time()
        while True:
            qes = xmhti.get_qspi_erase_status()
            if not qes.erase_in_progress:
                break

        t1 = time.time()
        print('Erase completed in %.2f seconds (ER Reg=0x%02X).' %
              (t1 - t0, qes.extended_read_register))

    # If requested, store a timestamp.
    if args.timestamp:
        t = xmhti.record_timestamp()
        print('Wrote timestamp %u.' % t)

    # If necessary, read the flash log.
    if args.read_flash_log:
        print('Reading flash log...')
        t0 = time.time()
        flash_log_data = xmhti.read_flash_log()
        dt = time.time() - t0
        if dt > 0:
            print('Read %u bytes in %.2f seconds (%.2f K/s).'
                  % (len(flash_log_data), dt, len(flash_log_data) / (1024*dt)))
        else:
            print('Read %u bytes.' % len(flash_log_data))

        with open(args.read_flash_log, 'wb') as f:
            f.write(flash_log_data)

    # If necessary, read the flash.
    if args.read_qspi or args.decode_qspi:
        print('Reading QSPI flash...')
        t0 = time.time()
        if args.read_qspi:
            qspi_data = xmhti.read_qspi_flash(0, qi.nslots)
        else:
            qspi_data = xmhti.read_qspi_flash(0, qi.write_index)
        dt = time.time() - t0
        if dt > 0:
            print('Read %u bytes in %.2f seconds (%.2f K/s).'
                  % (len(qspi_data), dt, len(qspi_data) / (1024*dt)))
        else:
            print('Read %u bytes.' % len(qspi_data))

    # If requested, write the flash contents to a file.
    if args.read_qspi:
        with open(args.read_qspi, 'wb') as f:
            f.write(qspi_data)

    # If requested, decode the flash contents.
    if args.decode_qspi:
        if not os.path.isdir(args.decode_qspi):
            raise Exception('Not a directory: %s' % args.decode_qspi)
        xmhti_qspi_decode.process_image(xmhti.serial_num, args.decode_qspi,
                                        qspi_data, bs.hse_freq, 0,
                                        xmhti.poly_psi, xmhti.poly_temp)

    # If requested, clear the chip erase journal.
    if args.clear_erase_journal:
        print('Clearing chip-erase journal...')
        xmhti.clear_erase_journal()

    # If requested, clear the autonomous config journal.
    if args.clear_auto_journal:
        print('Clearing autonomous-config journal...')
        xmhti.clear_auto_journal()

    # If requested, clear the flash log.
    if args.clear_flash_log:
        print('Clearing flash log...')
        xmhti.clear_flash_log()

    # If requested, set the autonomous configuration.
    if args.set_autonomous_config:
        print('Setting autonomous configuration...')
        words = args.set_autonomous_config.split(',')
        assert len(words) == 2
        interval_secs = int(words[0])
        flags         = int(words[1], 0)
        xmhti.set_autonomous_config(interval_secs, flags)

    # If requested, start continuous measurement mode.
    if args.measure_continuous:
        measure_continuous(xmhti)

    # If requested, start autonomous measurement mode.
    if args.measure_autonomous:
        print('Starting autonomous mode...')
        xmhti.start_autonomous_mode()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--serial-number', '-s')
    parser.add_argument('--chip-erase-start', action='store_true')
    parser.add_argument('--chip-erase-wait-done', action='store_true')
    parser.add_argument('--timestamp', action='store_true')
    parser.add_argument('--read-qspi')
    parser.add_argument('--read-flash-log')
    parser.add_argument('--decode-qspi')
    parser.add_argument('--clear-erase-journal', action='store_true')
    parser.add_argument('--clear-auto-journal', action='store_true')
    parser.add_argument('--clear-flash-log', action='store_true')
    parser.add_argument('--measure-continuous', action='store_true')
    parser.add_argument('--measure-autonomous', action='store_true')
    parser.add_argument('--set-autonomous-config', help=('Set the autonomous '
                        'config: <interval_secs>,<flags>'))
    main(parser.parse_args())


if __name__ == '__main__':
    _main()



# Copyright (c) 2020 Phase Advanced Sensor Systems, Inc.
import sys
import math


FRAC_LUT = [
    '\u2500',
    '\u258F',
    '\u258E',
    '\u258D',
    '\u258C',
    '\u258B',
    '\u258A',
    '\u2589',
    ]


class piter:
    '''
    Iterator class that also displays a progress bar.
    '''
    def __init__(self, seq, width=40, verbose=True, clear_line=False):
        self.seq        = seq
        self.width      = width
        self.verbose    = verbose
        self.clear_line = clear_line
        self.end        = len(seq)
        self._iter      = iter(seq)
        self.v          = 0

    def __repr__(self):
        if self.end:
            frac, n_star = math.modf(self.v * self.width / self.end)
        else:
            frac, n_star = 0, self.width
        frac         = int(frac * 8)
        n_star       = int(n_star)
        n_space      = self.width - n_star - 1

        s = '\r\u2503' + '\u2588'*n_star
        if n_space >= 0:
            s += FRAC_LUT[frac] + '\u2500'*n_space
        s += ('\u2503 [%u / %u]' % (self.v, self.end))

        return s

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.verbose:
            sys.stdout.write(repr(self))

        try:
            self.v += 1
            return next(self._iter)
        except StopIteration:
            if self.verbose:
                if self.clear_line:
                    sys.stdout.write('\r\x1B[0K')
                else:
                    print()
            raise


def prange(*args, **kwargs):
    return piter(range(*args), **kwargs)



# Copyright (c) 2022 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import struct

import btype

from xtalx.tools.math import (PolynomialFit1D, PolynomialFit2D,
                              make_poly_bstruct)


CAL_F0_REFCLK           = (1 << 0)
CAL_F0_POLY_PSI         = (1 << 1)
CAL_F0_POLY_TEMP        = (1 << 2)
CAL_F0_COMP_STANDARD    = (1 << 3)
CAL_F0_REPORT_ID        = (1 << 4)
CAL_F0_OSC_STARTUP_MS   = (1 << 5)

CAL_FEATURES = [{CAL_F0_REFCLK         : 'MCU refclk',
                 CAL_F0_POLY_PSI       : 'PSI polynomial',
                 CAL_F0_POLY_TEMP      : 'Temperature polynomial',
                 CAL_F0_COMP_STANDARD  : 'Standard Comparators',
                 CAL_F0_REPORT_ID      : 'XtalxDB Calibration Report ID',
                 CAL_F0_OSC_STARTUP_MS : 'Oscillator startup period in ms',
                 },
                {},
                {},
                {},
                {},
                ]

CAL_SIG = 0x4C414358


def checksum_xor32(mem):
    assert (len(mem) % 4) == 0
    checksum = 0
    for offset in range(0, len(mem), 4):
        checksum ^= struct.unpack_from('<L', mem, offset)[0]
    return checksum


PolyPSI  = make_poly_bstruct(2, 25)
PolyTemp = make_poly_bstruct(1, 5)


class CalPage(btype.Struct):
    # Ensure this section is a multiple of 8 bytes.
    sig                 = btype.uint32_t(0xFFFFFFFF)
    checksum_xor32      = btype.uint32_t(0xFFFFFFFF)
    len                 = btype.uint32_t(0xFFFFFFFF)
    features            = btype.Array(btype.uint32_t(0xFFFFFFFF), 5)
    refclk_x10          = btype.uint32_t(0xFFFFFFFF)
    report_id           = btype.uint32_t(0xFFFFFFFF)
    poly_psi            = PolyPSI()
    poly_temp           = PolyTemp()
    osc_startup_time_ms = btype.uint32_t(0xFFFFFFFF)
    pad                 = btype.Array(btype.uint32_t(0xFFFFFFFF), 1)

    # Reserved area, should also be a multiple of 8 bytes.
    rsrv2               = btype.Array(btype.uint32_t(0xFFFFFFFF), 424)
    _EXPECTED_SIZE      = 2048

    @staticmethod
    def get_short_size():
        return CalPage._EXPECTED_SIZE - CalPage._TYPE_MAP['rsrv2']._N * 4

    def is_valid(self):
        if self.sig != CAL_SIG:
            return False
        if self.len > CalPage._EXPECTED_SIZE:
            return False

        self.poly_psi.pack()
        self.poly_temp.pack()

        data = self.pack()
        if checksum_xor32(data[:self.len]) != 0:
            return False

        return True

    def is_same(self, other):
        return self.pack() == other.pack()

    def get_refclk_x10(self):
        if self.features[0] & CAL_F0_REFCLK:
            return self.refclk_x10
        return None

    def get_report_id(self):
        if self.features[0] & CAL_F0_REPORT_ID:
            return self.report_id
        return None

    def get_osc_startup_time_ms(self):
        if self.features[0] & CAL_F0_OSC_STARTUP_MS:
            return self.osc_startup_time_ms
        return None

    def get_polynomials(self):
        if self.features[0] & CAL_F0_POLY_PSI:
            poly_psi = PolynomialFit2D.from_poly_bstruct(self.poly_psi)
        else:
            poly_psi = None

        if self.features[0] & CAL_F0_POLY_TEMP:
            poly_temp = PolynomialFit1D.from_poly_bstruct(self.poly_temp)
        else:
            poly_temp = None

        return poly_psi, poly_temp



# Copyright (c) 2025 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
from enum import IntEnum

import btype

from .xti import Measurement
from .exception import XtalXException


CRC_0x9B_LUT = [
    0x00, 0x9B, 0xAD, 0x36, 0xC1, 0x5A, 0x6C, 0xF7,
    0x19, 0x82, 0xB4, 0x2F, 0xD8, 0x43, 0x75, 0xEE,
    0x32, 0xA9, 0x9F, 0x04, 0xF3, 0x68, 0x5E, 0xC5,
    0x2B, 0xB0, 0x86, 0x1D, 0xEA, 0x71, 0x47, 0xDC,
    0x64, 0xFF, 0xC9, 0x52, 0xA5, 0x3E, 0x08, 0x93,
    0x7D, 0xE6, 0xD0, 0x4B, 0xBC, 0x27, 0x11, 0x8A,
    0x56, 0xCD, 0xFB, 0x60, 0x97, 0x0C, 0x3A, 0xA1,
    0x4F, 0xD4, 0xE2, 0x79, 0x8E, 0x15, 0x23, 0xB8,
    0xC8, 0x53, 0x65, 0xFE, 0x09, 0x92, 0xA4, 0x3F,
    0xD1, 0x4A, 0x7C, 0xE7, 0x10, 0x8B, 0xBD, 0x26,
    0xFA, 0x61, 0x57, 0xCC, 0x3B, 0xA0, 0x96, 0x0D,
    0xE3, 0x78, 0x4E, 0xD5, 0x22, 0xB9, 0x8F, 0x14,
    0xAC, 0x37, 0x01, 0x9A, 0x6D, 0xF6, 0xC0, 0x5B,
    0xB5, 0x2E, 0x18, 0x83, 0x74, 0xEF, 0xD9, 0x42,
    0x9E, 0x05, 0x33, 0xA8, 0x5F, 0xC4, 0xF2, 0x69,
    0x87, 0x1C, 0x2A, 0xB1, 0x46, 0xDD, 0xEB, 0x70,
    0x0B, 0x90, 0xA6, 0x3D, 0xCA, 0x51, 0x67, 0xFC,
    0x12, 0x89, 0xBF, 0x24, 0xD3, 0x48, 0x7E, 0xE5,
    0x39, 0xA2, 0x94, 0x0F, 0xF8, 0x63, 0x55, 0xCE,
    0x20, 0xBB, 0x8D, 0x16, 0xE1, 0x7A, 0x4C, 0xD7,
    0x6F, 0xF4, 0xC2, 0x59, 0xAE, 0x35, 0x03, 0x98,
    0x76, 0xED, 0xDB, 0x40, 0xB7, 0x2C, 0x1A, 0x81,
    0x5D, 0xC6, 0xF0, 0x6B, 0x9C, 0x07, 0x31, 0xAA,
    0x44, 0xDF, 0xE9, 0x72, 0x85, 0x1E, 0x28, 0xB3,
    0xC3, 0x58, 0x6E, 0xF5, 0x02, 0x99, 0xAF, 0x34,
    0xDA, 0x41, 0x77, 0xEC, 0x1B, 0x80, 0xB6, 0x2D,
    0xF1, 0x6A, 0x5C, 0xC7, 0x30, 0xAB, 0x9D, 0x06,
    0xE8, 0x73, 0x45, 0xDE, 0x29, 0xB2, 0x84, 0x1F,
    0xA7, 0x3C, 0x0A, 0x91, 0x66, 0xFD, 0xCB, 0x50,
    0xBE, 0x25, 0x13, 0x88, 0x7F, 0xE4, 0xD2, 0x49,
    0x95, 0x0E, 0x38, 0xA3, 0x54, 0xCF, 0xF9, 0x62,
    0x8C, 0x17, 0x21, 0xBA, 0x4D, 0xD6, 0xE0, 0x7B,
]


def crc8(data, csum=0xFF):
    for v in data:
        csum = CRC_0x9B_LUT[(csum ^ v) & 0xFF]
    return csum


class SPIErrorCode(IntEnum):
    OK          = 0
    BAD_LENGTH  = 1
    BAD_CSUM    = 2


class FrequencyResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    status         = btype.uint8_t()
    pressure_hz    = btype.float64_t()
    temperature_hz = btype.float64_t()
    _EXPECTED_SIZE = 18


class ConversionResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    status         = btype.uint8_t()
    pressure_psi   = btype.float64_t()
    temperature_c  = btype.float64_t()
    _EXPECTED_SIZE = 18


class FixedResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    status         = btype.uint8_t()
    pressure_psi   = btype.int32_t()
    temperature_c  = btype.int32_t()
    _EXPECTED_SIZE = 10


class FullResponse(btype.Struct, endian='<'):
    age_ms         = btype.uint8_t()
    status         = btype.uint8_t()
    pressure_psi   = btype.float64_t()
    temperature_c  = btype.float64_t()
    pressure_hz    = btype.float64_t()
    temperature_hz = btype.float64_t()
    _EXPECTED_SIZE = 34


class ProtocolError(XtalXException):
    '''
    The protocol itself had garbage data.
    '''
    def __init__(self, tx_cmd, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tx_cmd = tx_cmd
        self.data = data

    def __str__(self):
        return 'ProtocolError(tx_cmd="%s", data="%s")' % (
                self.tx_cmd.hex(), self.data.hex())


class OpcodeMismatchError(XtalXException):
    '''
    The snesor thought it received a different command than the one we sent.
    '''
    def __init__(self, tx_cmd, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tx_cmd = tx_cmd
        self.data = data

    def __str__(self):
        return 'OpcodeMismatchError(tx_cmd="%s", data="%s")' % (
                self.tx_cmd.hex(), self.data.hex())


class RXChecksumError(XtalXException):
    '''
    We received a bad checksum in the response from the sensor.
    '''
    def __init__(self, tx_cmd, data, exp_csum, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tx_cmd = tx_cmd
        self.data = data
        self.exp_csum = exp_csum

    def __str__(self):
        return 'RXChecksumError(tx_cmd="%s", data="%s")' % (
                self.tx_cmd.hex(), self.data.hex())


class PrevCommandChecksumError(XtalXException):
    '''
    The previous command failed due to the sensor receiving a command with a
    bad checksum.
    '''


class PrevCommandUnrecognizedError(XtalXException):
    '''
    The previous command failed with a status code that we don't recognize.
    '''
    def __init__(self, err_code):
        super().__init__()
        self.err_code = err_code


class Comms:
    '''
    Communication protocol for sensor firmare 0.9.2 or later.
    '''
    def __init__(self, xhtiss):
        self.xhtiss = xhtiss

        self._synchronize()

        self.nop(corrupt_csum=1)
        assert self._read_err() == SPIErrorCode.BAD_CSUM

        try:
            self.nop()
        except PrevCommandChecksumError:
            pass
        except Exception as exc:
            raise Exception('Should have had a checksum error!') from exc
        assert self._read_err() == SPIErrorCode.OK

    def _csum_transact(self, cmd, corrupt_csum=0):
        tx_csum = crc8(cmd) + corrupt_csum
        tx_cmd  = cmd + bytes([tx_csum])
        data    = self.xhtiss.bus.transact(tx_cmd)
        if data[0] != 0xAA:
            raise ProtocolError(tx_cmd, data)
        if data[2] != cmd[0]:
            raise OpcodeMismatchError(tx_cmd, data)

        rsp      = data[:-1]
        exp_csum = crc8(rsp)
        if exp_csum != data[-1]:
            raise RXChecksumError(tx_cmd, data, exp_csum)

        return rsp[3:]

    def _read_err(self):
        tx_cmd = b'\x00\x00'
        data = self.xhtiss.bus.transact(tx_cmd)
        if data[0] != 0xAA:
            raise ProtocolError(tx_cmd, data)
        return data[1]

    def _synchronize(self):
        tx_cmd = b'\x34\x00\x00'
        tx_cmd = tx_cmd + bytes([crc8(tx_cmd)])
        while True:
            rsp = self.xhtiss.bus.transact(tx_cmd)
            if rsp[0] != 0xAA:
                continue
            if rsp[1] == 0xBB:
                raise Exception('Unsupported firmware version.')
            if rsp[1] != 0x00:
                continue
            if rsp[2] != 0x34:
                continue
            if rsp[3] != crc8(rsp[0:3]):
                continue

            rsp = self.xhtiss.bus.transact(b'\x00\x00')
            if rsp[0] != 0xAA:
                continue
            if rsp[1] != 0x00:
                continue

            return 0

    def _set_nvstore(self, addr, data):
        assert len(data) == 4
        cmd = bytes([0x1E, 0x00, 0x00, (addr & 0xFF), ((addr >> 8) & 0xFF),
                     data[0], data[1], data[2], data[3]])
        self._csum_transact(cmd)

    def _get_nvstore(self, addr, size):
        cmd = bytes([0x2A, 0x00, 0x00, (addr & 0xFF), ((addr >> 8) & 0xFF),
                     0x00])
        cmd += bytes(size)
        data = self._csum_transact(cmd)
        return data[3:]

    def exec_cmd(self, cmd, rsp_len):
        cmd_bytes = bytes([cmd, 0x00, 0x00]) + b'\x00'*rsp_len
        return self._csum_transact(cmd_bytes)

    def nop(self, corrupt_csum=0):
        self._csum_transact(b'\x34\x00\x00', corrupt_csum=corrupt_csum)

    def read_frequencies(self):
        data = self.exec_cmd(0x19, FrequencyResponse._STRUCT.size)
        return FrequencyResponse.unpack(data)

    def read_conversion(self):
        data = self.exec_cmd(0x07, ConversionResponse._STRUCT.size)
        return ConversionResponse.unpack(data)

    def read_fixed(self):
        data = self.exec_cmd(0x33, FixedResponse._STRUCT.size)
        return FixedResponse.unpack(data)

    def read_full(self):
        data = self.exec_cmd(0x2D, FullResponse._STRUCT.size)
        return FullResponse.unpack(data)

    def read_measurement(self):
        rsp = self.read_full()

        m = Measurement(self.xhtiss, None, rsp.pressure_psi, rsp.temperature_c,
                        rsp.pressure_hz, rsp.temperature_hz, None, None, None,
                        None, None, None, None, None, None)
        m._age_ms = rsp.age_ms
        m._status = rsp.status
        return m



# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.


def crc16_byte_reversed(v, P):
    for _ in range(8):
        v = (v >> 1) ^ (P * (v & 1))
    return v & 0xFFFF


LUT = [crc16_byte_reversed(i, 0xA001) for i in range(256)]


def compute(data, v=0xFFFF):
    for d in data:
        v = (v >> 8) ^ LUT[(v ^ d) & 0xFF]
    return v


def compute_as_bytes(data, **kwargs):
    crc = compute(data, **kwargs)
    return bytes([crc & 0xFF, (crc >> 8) & 0xFF])


assert compute(bytes([0x01, 0x04, 0x08, 0x00, 0x00, 0x00, 0x09, 0x00,
                      0x00, 0x00, 0x00])) == 0x0CF8



# Copyright (c) 2026 Phase Advanced Sensor Systems, Inc.
from .usbcmd import Status, Command, Response, CommandException, Device


__all__ = [
    'Command',
    'CommandException',
    'Device',
    'Response',
    'Status',
]



# Copyright (c) 2025 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import time
import logging

import xtalx.spi_adapter

from . import xhtiss_091
from . import xhtiss_092
from .cal_page import CalPage


class XHTISS:
    '''
    This is a driver for the high-temperature XHTISS sensor with an SPI
    interface.
    '''
    def __init__(self, bus):
        self.bus               = bus
        self._halt_yield       = True
        self.last_time_ns      = 0
        self.poll_interval_sec = 0

        self.comms = self._probe_comms()

        (self.serial_num,
         self.fw_version_str,
         self.fw_version,
         self.git_sha1) = self._read_ids()

        self.cal_page = self.read_valid_calibration_page()

        self.report_id = None
        self.poly_psi  = None
        self.poly_temp = None
        if self.cal_page is not None:
            self.report_id = self.cal_page.get_report_id()
            self.poly_psi, self.poly_temp = self.cal_page.get_polynomials()

    def __str__(self):
        return 'XHTISS(%s)' % self.serial_num

    def _probe_comms(self):
        # Probe the target firmware.
        tx_cmd = b'\x34\x00\x00'
        tx_cmd = tx_cmd + bytes([xhtiss_092.crc8(tx_cmd)])
        retries = 0
        while True:
            # Send a NOP command.  This will be interpreted as an unsupported
            # command by 0.9.1 firmware.
            rsp = self.bus.transact(tx_cmd)
            if rsp == b'????':
                retries += 1
                if retries == 10:
                    logging.info('Repeated probe rsp: %s, 0.9.1 firmware '
                                 'likely dead', rsp)
                    raise xhtiss_091.DeadFirmwareException()
                continue
            if rsp == b'\xAA\xBB??':
                return xhtiss_091.Comms(self)

            # The response didn't look like 0.9.1 firmware, analyze it.
            if rsp[:3] != b'\xAA\x00\x34':
                logging.info('Response prefix mismatch probing comms, '
                             'expected: %s rsp: %s', b'\xAA\x00\x34', rsp)
                time.sleep(1)
                continue
            csum = xhtiss_092.crc8(rsp[0:3])
            if rsp[3] != csum:
                logging.info('Response checksum error probing comms, '
                             'expected: 0x%02X rsp: %s', csum, rsp)
                time.sleep(1)
                continue

            # The response was good for an 0.9.2 or higher firmware; we can
            # double-check the sticky status code.
            rsp = self.bus.transact(b'\x00\x00')
            if rsp[:2] != b'\xAA\x00':
                logging.info('Response prefix mismatch probing sticky status, '
                             'expected: %s rsp: %s', b'\xAA\x00', rsp)
                time.sleep(1)
                continue

            return xhtiss_092.Comms(self)

    def _read_ids(self):
        data = self.comms._get_nvstore(0xCA01, 24)
        if data[0] == 0xFF or data[0] == 0x00:
            raise Exception('Invalid ID response from sensor, may not be '
                            'connected or powered.')
        serial_number = data.decode().strip('\x00')

        data = self.comms._get_nvstore(0xCA02, 10)
        fw_version_str = data.decode().strip('\x00')
        parts          = fw_version_str.split('.')
        fw_version = ((int(parts[0]) << 8) |
                      (int(parts[1]) << 4) |
                      (int(parts[2]) << 0))

        data = self.comms._get_nvstore(0xCA03, 48)
        git_sha1 = data.decode().strip('\x00')

        return serial_number, fw_version_str, fw_version, git_sha1

    def get_flash_params(self):
        data = self.comms._get_nvstore(0xC0EF, 4)
        t_c = data[0]
        p_c = data[1]
        sample_ms = (data[3] << 8) | (data[2] << 0)
        return t_c, p_c, sample_ms

    def set_flash_params(self, t_c, p_c, sample_ms):
        data = bytes([t_c, p_c, sample_ms & 0xFF, (sample_ms >> 8) & 0xFF])
        self.comms._set_nvstore(0xC0EF, data)

    def read_calibration_pages_raw(self):
        '''
        Returns the raw data bytes for the single calibration page stored in
        flash.
        '''
        data = b''
        for i in range(CalPage.get_short_size() // 4):
            address = 0x2000 + i*4
            data += self.comms._get_nvstore(address, 4)
        pad = b'\xff' * (CalPage._EXPECTED_SIZE - len(data))
        return (data + pad,)

    def read_calibration_pages(self):
        '''
        Returns a CalPage struct for the single calibration page in sensor
        flash, even if the page is missing or corrupt.
        '''
        (cp_data,) = self.read_calibration_pages_raw()
        cp = CalPage.unpack(cp_data)
        return (cp,)

    def read_valid_calibration_page(self):
        '''
        Returns CalPage struct from the sensor flash.  Returns None if the
        calibration is not present or corrupted.
        '''
        (cp,) = self.read_calibration_pages()
        return cp if cp.is_valid() else None

    def yield_measurements(self, poll_interval_sec=None, **_kwargs):
        if poll_interval_sec is None:
            poll_interval_sec = self.poll_interval_sec

        self._halt_yield = False
        while not self._halt_yield:
            try:
                m = self.comms.read_measurement()
            except xhtiss_092.OpcodeMismatchError as e:
                logging.info('%s: Opcode mismatch: tx_cmd "%s" data "%s"',
                             self, e.tx_cmd.hex(), e.data.hex())
                continue
            except xhtiss_092.RXChecksumError as e:
                logging.info('%s: RX checksum error: tx_cmd "%s" data "%s" '
                             'exp 0x%02X', self, e.tx_cmd.hex(), e.data.hex(),
                             e.exp_csum)
                continue
            except Exception as e:
                logging.info('%s: Exception: %s', self, e)
                raise
            if m._age_ms > 25:
                continue
            if isinstance(self.bus, xtalx.spi_adapter.SPIA):
                m._current_amps = self.bus.measure_current()

            yield m
            time.sleep(poll_interval_sec)

    def halt_yield(self):
        self._halt_yield = True

    def time_ns_increasing(self):
        '''
        Returns a time value in nanoseconds that is guaranteed to increase
        after every single call.  This function is not thread-safe.
        '''
        self.last_time_ns = t = max(time.time_ns(), self.last_time_ns + 1)
        return t



# Copyright (c) 2022 by Phase Advanced Sensor Systems.
# All rights reserved.
from datetime import datetime, timezone
import argparse
import struct
import math
import os

import numpy as np


SS_NOP              = 0
SS_TEXT             = 1
SS_USB_STARTUP      = 2
SS_BATTERY_STARTUP  = 3
SS_TIMESTAMP        = 4
SS_AUTONOMOUS_START = 5
SS_PT_FREQS         = 6
SS_TPLC_COUNTS      = 7
SS_TPL_FREQS        = 8

BLANK_SLOT = b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'


class Record:
    def __init__(self, data):
        self.data = data


class BlankRecord(Record):
    '''
    8 bytes of 0xFF.
    '''
    def __repr__(self):
        return 'BlankRecord: FFFFFFFFFFFFFFFF'


class JunkRecord(Record):
    '''
    Data that cannot be decoded.
    '''
    def __repr__(self):
        return 'JunkRecord: %s' % self.data.hex()


class NopRecord(Record):
    def __repr__(self):
        return 'NOP: %s' % self.data.hex()


class USBStartup(Record):
    def __init__(self, data):
        super().__init__(data)
        (self.post_result,
         self.csr,
         self.sr1,
         self.reset_reason) = struct.unpack('<IIIB', data)

    def __repr__(self):
        return ('USBStartup: POST=0x%08X CSR=0x%08X SR1=0x%08X RR=0x%02X' %
                (self.post_result, self.csr, self.sr1, self.reset_reason))


class Timestamp(Record):
    def __init__(self, data):
        super().__init__(data)
        self.timestamp = struct.unpack_from('<I', data)[0]

    def __repr__(self):
        dt  = datetime.fromtimestamp(self.timestamp, timezone.utc)
        ldt = dt.astimezone(tz=None)
        return 'Timestamp: %s' % ldt.strftime('%c %Z')


class AutoStart(Record):
    def __init__(self, data):
        super().__init__(data)
        self.period_sec = struct.unpack_from('<I', data)[0]

    def __repr__(self):
        return 'AutoStart: %u sec interval' % self.period_sec


class TPLC(Record):
    def __init__(self, data, ref_freq, bias_count):
        super().__init__(data)
        self.t_count = (data[0] << 16) | (data[ 1] << 8) | data[ 2]
        self.p_count = (data[3] << 16) | (data[ 4] << 8) | data[ 5]
        self.l_count = (data[6] << 16) | (data[ 7] << 8) | data[ 8]
        self.c_count = (data[9] << 16) | (data[10] << 8) | data[11]
        self.iter    = data[12]
        self.csum    = data[13]

        self.t_hz = self.p_hz = self.l_hz = 0

        if self.t_count != 0xFFFFFF:
            self.t_count += bias_count
            self.t_hz     = ref_freq * 26200 / self.t_count
        else:
            self.t_count = 0xFFFFFFFF
        if self.p_count != 0xFFFFFF:
            self.p_count += bias_count
            self.t_hz     = ref_freq * 5000 / self.p_count
        else:
            self.p_count = 0xFFFFFFFF
        if self.l_count != 0xFFFFFF:
            self.l_count += bias_count
            self.l_hz     = ref_freq * 3272 / self.l_count
        else:
            self.l_count = 0xFFFFFFFF
        if self.c_count != 0xFFFFFF:
            self.c_count += bias_count
        else:
            self.c_count = 0xFFFFFFFF

    def __repr__(self):
        return ('TPLC(0x%02X): T%08X P%08X L%08X C%08X' %
                (self.iter, self.t_count, self.p_count, self.l_count,
                 self.c_count))


class PT(Record):
    def __init__(self, data):
        super().__init__(data)
        p_freq_csum = struct.unpack('<I', data[0:4])[0]
        self.p_freq = (p_freq_csum & 0x0FFFFFFF)
        self.csum   = (p_freq_csum >> 28)
        self.t_freq = (struct.unpack('<I', data[3:7])[0] >> 8)
        if self.p_freq in (0, 0x0FFFFFFF):
            self.p_hz = math.nan
        else:
            self.p_hz = 37000 + (16384 * self.p_freq / 2**28)
        if self.t_freq in (0, 0x00FFFFFF):
            self.t_hz = math.nan
        else:
            self.t_hz = 260952 + (4096 * self.t_freq / 2**24)

    def __repr__(self):
        return ('PT: T%.6f  P%.6f' % (self.t_hz, self.p_hz))


class TPL(Record):
    def __init__(self, data):
        super().__init__(data)
        self.rt_freq = struct.unpack('<I', data[0:4])[0]
        self.rp_freq = struct.unpack('<I', data[4:8])[0]
        self.rl_freq = struct.unpack('<I', data[8:12])[0]
        if self.rt_freq in (0, 0xFFFFFFFF):
            self.t_hz = math.nan
        else:
            self.t_hz = 260952 + (4096 * self.rt_freq / 2**32)
        if self.rp_freq in (0, 0xFFFFFFFF):
            self.p_hz = math.nan
        else:
            self.p_hz = 37000 + (16384 * self.rp_freq / 2**32)
        if self.rl_freq in (0, 0xFFFFFFFF):
            self.l_hz = math.nan
        else:
            self.l_hz = 31744 + (2048 * self.rl_freq / 2**32)
        self.iter    = data[12]
        self.csum    = data[13]

    def __repr__(self):
        return ('TPL(0x%02X): T%.6f  P%.6f  L%.6f' %
                (self.iter, self.t_hz, self.p_hz, self.l_hz))


class Unknown(Record):
    def __init__(self, op, data):
        super().__init__(data)
        self.op = op

    def __repr__(self):
        return 'Unknown(%2u): %s' % (self.op, self.data.hex())


def parse_record(data, ref_freq, bias_count):
    '''
    Parse a sequence of 8-byte slots into a record.
    '''
    record_data = b''
    for i in range(len(data) // 8):
        record_data += data[i*8 + 1:i*8 + 8]

    op = data[0] & 0x1F
    if op == SS_NOP:
        return NopRecord(record_data)
    if op == SS_USB_STARTUP and len(data) == 16:
        return USBStartup(record_data)
    if op == SS_TIMESTAMP and len(data) == 8:
        return Timestamp(record_data)
    if op == SS_AUTONOMOUS_START and len(data) == 8:
        return AutoStart(record_data)
    if op == SS_TPLC_COUNTS and len(data) == 16:
        return TPLC(record_data, ref_freq, bias_count)
    if op == SS_PT_FREQS and len(data) == 8:
        return PT(record_data)
    if op == SS_TPL_FREQS and len(data) == 16:
        return TPL(record_data)
    return Unknown(op, record_data)


class QSPIDecoder:
    def __init__(self, image, ref_freq, bias_count):
        self.image      = image
        self.ref_freq   = ref_freq
        self.bias_count = bias_count
        self.pos        = 0

    def pop_slot_data(self):
        if len(self.image) - self.pos < 8:
            return None

        slot_data = self.image[self.pos:self.pos + 8]
        self.pos += 8
        return slot_data

    def pop_record(self):
        # Get an 8-byte slot.
        slot_data = self.pop_slot_data()
        if slot_data is None:
            return None
        if slot_data == BLANK_SLOT:
            return BlankRecord(slot_data)

        # Check the slot header for the First bit.
        hdr0 = slot_data[0]
        if not hdr0 & 0x80:
            print('Missing F')
            return JunkRecord(slot_data)

        # Pop records as necessary until we see Last or something unexpected.
        data = slot_data
        hdr  = hdr0
        while not hdr & 0x40:
            # Retrieve the next slot.
            slot_data = self.pop_slot_data()
            if slot_data is None:
                return JunkRecord(data)

            # Check if record is blank.
            if slot_data == BLANK_SLOT:
                print('Missing L')
                self.pos -= 8
                return JunkRecord(data)

            # Check if First bit is set.
            hdr = slot_data[0]
            if hdr & 0x80:
                print('Redundant F')
                self.pos -= 8
                return JunkRecord(data)

            # Check if OP has changed.
            if (hdr & 0x1F) != (hdr0 & 0x1F):
                print('OP changed')
                self.pos -= 8
                return JunkRecord(data)

            # The data belongs with this record, so append it.
            data += slot_data

        # The Last bit is set, the record is complete.
        return parse_record(data, self.ref_freq, self.bias_count)


def process_image(fname, directory, data, ref_freq, bias_count,
                  poly_psi=None, poly_temp=None):
    qd             = QSPIDecoder(data, ref_freq, bias_count)
    fname          = os.path.basename(fname)
    f_num          = 1
    auto_entry     = None
    tplc_entries   = []
    tpl_entries    = []
    pt_entries     = []
    blank_rollup   = None
    last_timestamp = None
    while True:
        r = qd.pop_record()

        if isinstance(r, TPLC):
            tplc_entries.append(r)
            continue

        if isinstance(r, TPL):
            tpl_entries.append(r)
            continue

        if isinstance(r, PT):
            pt_entries.append(r)
            continue

        if isinstance(r, NopRecord):
            continue

        if tplc_entries:
            period = 0 if auto_entry is None else auto_entry.period_sec
            f_name = os.path.join(directory, fname + '.tplc.%u.csv' % f_num)
            f_num += 1
            with open(f_name, 'w', encoding='utf8') as f:
                f.write('iter,t,t_count,p_count,l_count,c_count,'
                        't_hz,p_hz,l_hz\n')
                for i, tplc in enumerate(tplc_entries):
                    f.write('%u,%u,%u,%u,%u,%u,%.6f,%.6f,%.6f\n' %
                            (tplc.iter, i * period, tplc.t_count, tplc.p_count,
                             tplc.l_count, tplc.c_count, tplc.t_hz, tplc.p_hz,
                             tplc.l_hz))
            print('     TPLC: %u 16-byte entries written to %s' %
                  (len(tplc_entries), f_name))
            tplc_entries = []

        if tpl_entries:
            period = 0 if auto_entry is None else auto_entry.period_sec
            f_name = os.path.join(directory, fname + '.tpl.%u.csv' % f_num)
            f_num += 1
            with open(f_name, 'w', encoding='utf8') as f:
                f.write('iter,t,t_hz,p_hz,l_hz\n')
                for i, tpl in enumerate(tpl_entries):
                    f.write('%u,%u,%.6f,%.6f,%.6f\n' %
                            (tpl.iter, i * period, tpl.t_hz, tpl.p_hz,
                             tpl.l_hz))
            print('     TPL: %u 16-byte entries written to %s' %
                  (len(tpl_entries), f_name))
            tpl_entries = []

        if pt_entries:
            if auto_entry is not None:
                period = auto_entry.period_sec or 0.1
            else:
                period = 0
            f_name = os.path.join(directory, fname + '.pt.%u.csv' % f_num)
            f_num += 1
            with open(f_name, 'w', encoding='utf8') as f:
                hdrs = 't,t_hz,p_hz,temp_c,psi'
                f.write(hdrs + '\n')
                t_hzs = np.array([pt.t_hz for pt in pt_entries])
                p_hzs = np.array([pt.p_hz for pt in pt_entries])
                temp_cs = None
                psis = None
                if poly_temp is not None:
                    temp_cs = poly_temp(t_hzs)
                if poly_psi is not None:
                    psis = poly_psi(p_hzs, t_hzs)
                for i in range(len(pt_entries)):
                    t_hz   = t_hzs[i]
                    p_hz   = p_hzs[i]
                    temp_c = temp_cs[i] if temp_cs is not None else None
                    psi    = psis[i] if psis is not None else None
                    line = '%.1f,%.6f,%.6f' % (i * period, t_hz, p_hz)
                    if temp_c is not None:
                        line += ',%.4f' % temp_c
                    else:
                        line += ','
                    if psi is not None:
                        line += ',%.5f' % psi
                    else:
                        line += ','
                    f.write(line + '\n')
            print('     PT: %u 8-byte entries written to %s' %
                  (len(pt_entries), f_name))
            pt_entries = []

        if isinstance(r, AutoStart):
            auto_entry = r

        if isinstance(r, BlankRecord):
            blank_rollup = 1 if blank_rollup is None else blank_rollup + 1
            continue

        if blank_rollup is not None:
            print('    Blank: %u 8-byte entries' % blank_rollup)
            blank_rollup = None

        if r is None:
            break

        if isinstance(r, Timestamp):
            if last_timestamp is not None:
                dt = r.timestamp - last_timestamp.timestamp
                print('%s (+%u seconds)' % (r, dt))
            else:
                print(r)
            last_timestamp = r
        else:
            print(r)


def main(args):
    if not os.path.isdir(args.directory):
        raise Exception('Not a directory: %s' % args.directory)
    with open(args.file, 'rb') as f:
        data = f.read()
    process_image(args.file, args.directory, data, args.ref_freq,
                  args.bias_count)


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--file', '-f', required=True)
    parser.add_argument('--directory', '-d', required=True)
    parser.add_argument('--ref-freq', '-r', type=int, default=159600000)
    parser.add_argument('--bias-count', '-b', type=int, default=12053700)
    main(parser.parse_args())


if __name__ == '__main__':
    _main()



import argparse

import glotlib
import numpy as np

from xtalx.tools import csv


CSV_FORMAT = {
    't'             : float,
    't_hz'          : float,
    'p_hz'          : float,
    'temp_c'        : float,
    'psi'           : float,
}


def main(args):
    with open(args.csv, 'r', encoding='utf8') as f:
        d = csv.Decoder(f, CSV_FORMAT)

    w = glotlib.Window(512, 512, msaa=4)
    p_plot = w.add_plot(211)
    t_plot = w.add_plot(212, sharex=p_plot)

    PRESSURE = np.array([p['psi'] for p in d.points])
    if any(p is None for p in PRESSURE):
        p_plot.set_y_label('Pressure (Hz)')
        PRESSURE = np.array([p['p_hz'] for p in d.points])
    else:
        p_plot.set_y_label('Pressure (PSI)')

    TEMP = np.array([p['temp_c'] for p in d.points])
    if any(t is None for t in TEMP):
        t_plot.set_y_label('Temp (Hz)')
        TEMP = np.array([p['t_hz'] for p in d.points])
    else:
        t_plot.set_y_label('Temp (C)')
    t_plot.set_x_label('Time (seconds)')

    X = np.array([p['t'] for p in d.points])

    p_plot.add_steps(X=X, Y=PRESSURE)
    t_plot.add_steps(X=X, Y=TEMP)
    p_plot.snap_bounds()
    t_plot.snap_bounds()

    try:
        glotlib.interact()
    except KeyboardInterrupt:
        print()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--csv', required=True)
    main(parser.parse_args())


if __name__ == '__main__':
    _main()



# Copyright (c) 2026 Phase Advanced Sensor Systems, Inc.
import argparse
import threading
import time

import xtalx.spi_adapter


MONITORING_CURRENT = True


def make_spia(args):
    dev = xtalx.spi_adapter.find_one_spia(serial_number=args.serial_number)
    if dev is None:
        raise Exception('No SPIA found.')

    return xtalx.spi_adapter.make_spia(dev)


def current_monitor_thread(spia):
    print('Monitoring current consumption...')
    while MONITORING_CURRENT:
        amps = spia.measure_current()
        print('%.2f mA' % (amps * 1000))

        time.sleep(0.1)


def main(args):
    global MONITORING_CURRENT

    spia = make_spia(args)
    print('Found MBA: %s' % spia.serial_num)
    print('  FW SHA1: %s' % spia.git_sha1)
    print('FW Vesion: %s' % spia.fw_version_str)
    print('----------------------')

    print('Enabling VEXT...')
    spia.set_vext(True)

    cmt = threading.Thread(target=current_monitor_thread, args=(spia,))
    cmt.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        MONITORING_CURRENT = False

    cmt.join()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--serial-number', '-s')
    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()



# Copyright (c) 2023 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import threading
import argparse
import math
import time

import glotlib

import xtalx.p_sensor
import xtalx.modbus_adapter
import xtalx.spi_adapter
import xtalx.tools.modbus.serial
from xtalx.tools.math import XYSeries


LINE_WIDTH = 1


class TrackerWindow(glotlib.Window):
    def __init__(self, name, period, show_lores_data, display_frequencies):
        super().__init__(900, 700, msaa=4, name=name)

        self.period              = period
        self.show_lores_data     = show_lores_data
        self.display_frequencies = display_frequencies
        self.data_gen            = -1
        self.plot_gen            = -1
        self.data_lock           = threading.Lock()
        self.new_data            = []
        self.p_measurements      = XYSeries([], [])
        self.lp_measurements     = XYSeries([], [])

        if display_frequencies:
            self.p_plot = self.add_plot(
                311, limits=(-0.1, 47000, 120, 62000), max_v_ticks=10)
        else:
            self.p_plot = self.add_plot(
                311, limits=(-0.1, -0.5, 120, 300), max_v_ticks=10)
        self.lp_lines = self.p_plot.add_steps(X=[], Y=[], width=LINE_WIDTH)
        self.p_lines = self.p_plot.add_steps(X=[], Y=[], width=LINE_WIDTH)
        self.lp_lines.color = (0.78, 0.87, 0.93, 1)
        self.p_plot.set_y_label('P Hz' if display_frequencies else 'PSI')

        if display_frequencies:
            self.p_slow_plot = self.add_plot(
                312, limits=(-0.1, 47000, 120, 62000), max_v_ticks=10,
                sharex=self.p_plot, sharey=self.p_plot)
        else:
            self.p_slow_plot = self.add_plot(
                312, limits=(-0.1, -0.5, 120, 300), max_v_ticks=10,
                sharex=self.p_plot, sharey=self.p_plot)
        self.lp_slow_lines = self.p_slow_plot.add_lines(X=[], Y=[],
                                                        width=LINE_WIDTH)
        self.p_slow_lines = self.p_slow_plot.add_lines(X=[], Y=[],
                                                       width=LINE_WIDTH)
        self.p_slow_plot.set_y_label(
            ('P Hz (%u-sec Avg)' % period) if display_frequencies else
            ('PSI (%u-sec Avg)' % period))

        if display_frequencies:
            self.t_plot = self.add_plot(
                313, limits=(-0.1, 260000, 120, 264000), max_v_ticks=10,
                sharex=self.p_plot)
        else:
            self.t_plot = self.add_plot(
                313, limits=(-0.1, -0.5, 120, 50), max_v_ticks=10,
                sharex=self.p_plot)
        self.lt_lines = self.t_plot.add_steps(X=[], Y=[], width=LINE_WIDTH)
        self.t_lines = self.t_plot.add_steps(X=[], Y=[],  width=LINE_WIDTH)
        self.lt_lines.color = (0.78, 0.87, 0.93, 1)
        self.t_plot.set_y_label('Temp Hz' if display_frequencies else
                                'Temp (C)')

        self.pos_label = self.add_label((0.99, 0.01), '', anchor='SE')

        self.mouse_vlines = [self.p_plot.add_vline(0, color='#80C080'),
                             self.p_slow_plot.add_vline(0, color='#80C080'),
                             self.t_plot.add_vline(0, color='#80C080')]

        label_font = glotlib.fonts.vera_bold(48, 0)
        self.psi_label      = self.add_label(self.p_plot.bounds[2:4], '',
                                             anchor='NE', font=label_font)
        self.psi_slow_label = self.add_label(self.p_slow_plot.bounds[2:4], '',
                                             anchor='NE', font=label_font)
        self.temp_label     = self.add_label(self.t_plot.bounds[2:4], '',
                                             anchor='NE', font=label_font)

    def handle_mouse_moved(self, x, y):
        data_x, _ = self.p_plot._window_to_data(x, y)
        for vline in self.mouse_vlines:
            vline.set_x_data(data_x)
        self.mark_dirty()

    def update_geometry(self, _t):
        updated = False

        _, _, _, data_x, data_y = self.get_mouse_pos()
        if data_x is not None:
            updated |= self.pos_label.set_text('%.10f  %.10f' %
                                               (data_x, data_y))

        new_data = None
        with self.data_lock:
            if self.new_data:
                new_data = self.new_data
                self.new_data = []

        if new_data:
            updated = True

            # Low-res temperature measurements.
            if self.display_frequencies:
                X = [m._timestamp for m in new_data
                     if m.lores_temp_freq is not None]
                Y = [m.lores_temp_freq for m in new_data
                     if m.lores_temp_freq is not None]
            else:
                X = [m._timestamp for m in new_data
                     if m.lores_temp_c is not None]
                Y = [m.lores_temp_c for m in new_data
                     if m.lores_temp_c is not None]
            if self.show_lores_data:
                self.lt_lines.append_x_y_data(X, Y)

            # Hi-res temperature measurements.
            if self.display_frequencies:
                self.t_lines.append_x_y_data(
                    [m._timestamp for m in new_data],
                    [m.temp_freq for m in new_data])
                self.temp_label.set_text('%.4f Hz' % new_data[-1].temp_freq)
            else:
                self.t_lines.append_x_y_data(
                    [m._timestamp for m in new_data],
                    [m.temp_c for m in new_data])
                self.temp_label.set_text('%.4f \u00B0C' % new_data[-1].temp_c)

            # Low-res pressure (LP) measurements.
            if self.display_frequencies:
                X = [m._timestamp for m in new_data
                     if m.lores_pressure_freq is not None]
                Y = [m.lores_pressure_freq for m in new_data
                     if m.lores_pressure_freq is not None]
            else:
                X = [m._timestamp for m in new_data
                     if m.lores_pressure_psi is not None]
                Y = [m.lores_pressure_psi for m in new_data
                     if m.lores_pressure_psi is not None]
            lp_len = len(self.lp_slow_lines.vertices)
            if lp_len:
                lp_timestamp = self.lp_measurements.X[-1]
            self.lp_measurements.append(X, Y)
            if self.show_lores_data:
                self.lp_lines.append_x_y_data(X, Y)

            # Averaged data from LP measurements.
            if len(X):
                if lp_len:
                    t0    = int(lp_timestamp // self.period) * self.period
                    index = lp_len - 1
                else:
                    t0    = int(X[0] // self.period) * self.period
                    index = 0

                timestamps = []
                pressures  = []
                t          = t0
                while t <= X[-1]:
                    p = self.lp_measurements.get_avg_value(t, t + self.period)
                    if p is not None:
                        timestamps.append(t + self.period / 2)
                        pressures.append(p)
                    t += self.period
                if self.show_lores_data:
                    self.lp_slow_lines.sub_x_y_data(index, timestamps,
                                                    pressures)

            # Hi-res pressure (P) measurements.
            if self.display_frequencies:
                X = [m._timestamp for m in new_data]
                Y = [m.pressure_freq for m in new_data]
                self.psi_label.set_text('%.4f Hz' % new_data[-1].pressure_freq)
            else:
                X = [m._timestamp for m in new_data]
                Y = [m.pressure_psi for m in new_data]
                self.psi_label.set_text('%.4f PSI' % new_data[-1].pressure_psi)
            p_len = len(self.p_slow_lines.vertices)
            if p_len:
                p_timestamp = self.p_measurements.X[-1]
            self.p_measurements.append(X, Y)
            self.p_lines.append_x_y_data(X, Y)

            # Averaged data from P measurements.
            if p_len:
                t0    = int(p_timestamp // self.period) * self.period
                index = p_len - 1
            else:
                t0    = int(X[0] // self.period) * self.period
                index = 0

            timestamps = []
            pressures  = []
            t          = t0
            while t <= X[-1]:
                p = self.p_measurements.get_avg_value(t, t + self.period)
                if p is not None:
                    timestamps.append(t + self.period / 2)
                    pressures.append(p)
                t += self.period
            self.p_slow_lines.sub_x_y_data(index, timestamps, pressures)
            if len(pressures) >= 2:
                if self.display_frequencies:
                    self.psi_slow_label.set_text('%.4f Hz' % pressures[-2])
                else:
                    self.psi_slow_label.set_text('%.4f PSI' % pressures[-2])

        return updated

    def measurement_callback(self, m):
        with self.data_lock:
            self.new_data.append(m)
            self.mark_dirty()


def measure_thread(x, tw, csv_file, display_frequencies):
    t0 = time.time()
    for m in x.yield_measurements(do_reset=False):
        t = time.time()
        m._timestamp = dt = t - t0
        tw.measurement_callback(m)

        if csv_file:
            if display_frequencies:
                temp = m.temp_freq if m.temp_freq is not None else math.nan
                pressure = (m.pressure_freq if m.pressure_freq is not None
                            else math.nan)
            else:
                temp = m.temp_c if m.temp_c is not None else math.nan
                pressure = (m.pressure_psi if m.pressure_psi is not None
                            else math.nan)
            csv_file.write('%.6f,%.6f,%.5f,%.5f\n' % (t, dt, temp, pressure))
            csv_file.flush()


def make_sensor(args):
    if args.intf:
        bus = xtalx.tools.modbus.serial.Bus(args.intf, args.baud_rate)
        return xtalx.p_sensor.XHTISM(bus, int(args.modbus_addr, 0))

    dev = xtalx.p_sensor.find_one_xti(serial_number=args.serial_number)
    if dev is not None:
        return xtalx.p_sensor.make(dev)

    dev = xtalx.modbus_adapter.find_one_mba(serial_number=args.serial_number)
    if dev is not None:
        bus = xtalx.modbus_adapter.make_mba(dev, baud_rate=args.baud_rate)
        bus.set_vext(True)
        time.sleep(0.1)
        return xtalx.p_sensor.XHTISM(bus, int(args.modbus_addr, 0))

    dev = xtalx.spi_adapter.find_one_spia(serial_number=args.serial_number)
    if dev is not None:
        bus = xtalx.spi_adapter.make_spia(dev)
        bus.set_vext(True)
        time.sleep(0.1)
        return xtalx.p_sensor.XHTISS(bus)

    raise Exception('No matching devices.')


def main(args):
    x = make_sensor(args)

    if args.csv_file:
        if args.display_frequencies:
            hdr = 'time,dt,temp_hz,pressure_hz\n'
        else:
            hdr = 'time,dt,temp_c,pressure_psi\n'
        csv_file = open(  # pylint: disable=R1732
            args.csv_file, 'a+', encoding='utf8')

        pos = csv_file.tell()
        if pos != 0:
            csv_file.seek(0)
            if csv_file.read(28) != hdr:
                print('%s does not appear to be a pressure sensor log file.' %
                      args.csv_file)
                return
            csv_file.seek(pos)
        else:
            csv_file.write(hdr)
            csv_file.flush()
    else:
        csv_file = None

    tw  = TrackerWindow(x.serial_num, args.averaging_period_secs,
                        args.show_lores_data, args.display_frequencies)
    mt  = threading.Thread(target=measure_thread,
                           args=(x, tw, csv_file, args.display_frequencies))
    mt.start()

    try:
        glotlib.interact()
    except KeyboardInterrupt:
        print()
    finally:
        x.halt_yield()
        mt.join()


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--intf', '-i')
    parser.add_argument('--baud-rate', type=int, default=115200)
    parser.add_argument('--modbus-addr', '-m', default='0x80')
    parser.add_argument('--serial_number', '-s')
    parser.add_argument('--csv-file')
    parser.add_argument('--averaging-period-secs', type=int, default=3)
    parser.add_argument('--show-lores-data', action='store_true')
    parser.add_argument('--display-frequencies', action='store_true')
    args = parser.parse_args()
    main(args)


if __name__ == '__main__':
    _main()
